import os
import sys
from pathlib import Path

import pytest

from pipenv.utils.processes import subprocess_run

# This module is run only on Windows.
pytestmark = pytest.mark.skipif(os.name != "nt", reason="only relevant on windows")


@pytest.mark.project
def test_case_changes_windows(pipenv_instance_pypi):
    """Test project matching for case changes on Windows."""
    with pipenv_instance_pypi() as p:
        c = p.pipenv("install pytz")
        assert c.returncode == 0

        # Canonical venv location.
        c = p.pipenv("--venv")
        assert c.returncode == 0
        virtualenv_location = c.stdout.strip()

        # Dance around to change the casing of the project directory.
        target = p.path.upper()
        if target == p.path:
            target = p.path.lower()
        os.chdir("..")
        os.chdir(target)
        assert os.path.abspath(os.curdir) != p.path

        # Ensure the incorrectly-cased project can find the correct venv.
        c = p.pipenv("--venv")
        assert c.returncode == 0
        assert c.stdout.strip().lower() == virtualenv_location.lower()


@pytest.mark.files
@pytest.mark.local
def test_local_path_windows(pipenv_instance_pypi):
    whl = Path(__file__).parent.parent.joinpath(
        "pypi", "six", "six-1.11.0-py2.py3-none-any.whl"
    )
    try:
        whl = whl.resolve()
    except OSError:
        whl = whl.absolute()
    with pipenv_instance_pypi() as p:
        c = p.pipenv(f'install "{whl}" -v')
        assert c.returncode == 0


@pytest.mark.local
@pytest.mark.files
def test_local_path_windows_forward_slash(pipenv_instance_pypi):
    whl = Path(__file__).parent.parent.joinpath(
        "pypi", "six", "six-1.11.0-py2.py3-none-any.whl"
    )
    try:
        whl = whl.resolve()
    except OSError:
        whl = whl.absolute()
    with pipenv_instance_pypi() as p:
        c = p.pipenv(f'install "{whl.as_posix()}" -v')
        assert c.returncode == 0


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info[:2] == (3, 8),
    reason="Seems to work on 3.8 but not via the CI",
)
@pytest.mark.cli
def test_pipenv_clean_windows(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        c = p.pipenv("install dataclasses-json")
        assert c.returncode == 0
        c = p.pipenv(f"run pip install -i {p.index_url} click")
        assert c.returncode == 0

        c = p.pipenv("clean --dry-run")
        assert c.returncode == 0
        assert "click" in c.stdout.strip()


@pytest.mark.cli
def test_pipenv_run_with_special_chars_windows(pipenv_instance_pypi):
    with pipenv_instance_pypi():
        c = subprocess_run(["pipenv", "run", "echo", "[3-1]"])
        assert c.returncode == 0, c.stderr



import pytest

from .conftest import DEFAULT_PRIVATE_PYPI_SERVER


@pytest.mark.urls
@pytest.mark.extras
@pytest.mark.install
def test_install_uri_with_extras(pipenv_instance_pypi):
    server = DEFAULT_PRIVATE_PYPI_SERVER.replace("/simple", "")
    file_uri = f"{server}/packages/plette/plette-0.2.2-py2.py3-none-any.whl"
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = false
name = "testindex"

[packages]
plette = {{file = "{file_uri}", extras = ["validation"]}}
"""
            f.write(contents)
        c = p.pipenv("install")
        assert c.returncode == 0
        assert "plette" in p.lockfile["default"]
        assert "cerberus" in p.lockfile["default"]


@pytest.mark.star
@pytest.mark.install
def test_install_major_version_star_specifier(pipenv_instance_pypi):
    """Test that major version star specifiers like '1.*' work correctly."""
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = true
name = "pypi"

[packages]
six = "==1.*"
"""
            f.write(contents)
        c = p.pipenv("install")
        assert c.returncode == 0
        assert "six" in p.lockfile["default"]


@pytest.mark.star
@pytest.mark.install
def test_install_full_wildcard_specifier(pipenv_instance_pypi):
    """Test that full wildcard specifiers '*' work correctly."""
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = true
name = "pypi"

[packages]
requests = "*"
"""
            f.write(contents)
        c = p.pipenv("install")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]


@pytest.mark.star
@pytest.mark.install
def test_install_single_equals_star_specifier(pipenv_instance_pypi):
    """Test that single equals star specifiers like '=8.*' work correctly."""
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = true
name = "pypi"

[packages]
requests = "==2.*"
"""
            f.write(contents)
        c = p.pipenv("install")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]
        assert p.lockfile["default"]["requests"]["version"].startswith("==2.")


@pytest.mark.star
@pytest.mark.install
def test_install_command_with_star_specifier(pipenv_instance_pypi):
    """Test that star specifiers work when used in the install command."""
    with pipenv_instance_pypi() as p:
        # Initialize pipfile first
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = true
name = "pypi"

[packages]
"""
            f.write(contents)

        # Test with single equals and star specifier
        c = p.pipenv("install urllib3==1.*")
        assert c.returncode == 0
        assert "urllib3" in p.lockfile["default"]




from __future__ import annotations

import os
import re
import subprocess
from builtins import TimeoutError
from typing import TYPE_CHECKING, Any

from pipenv.vendor.packaging.version import InvalidVersion

if TYPE_CHECKING:
    from pathlib import Path

from ..exceptions import InvalidPythonVersion

# Regular expression for parsing Python version strings
version_re_str = (
    r"(?P<major>\d+)(?:\.(?P<minor>\d+))?(?:\.(?P<patch>(?<=\.)[0-9]+))?\.?"
    r"(?:(?P<prerel>[abc]|rc)(?:(?P<prerelversion>\d+(?:\.\d+)*))?)"
    r"?(?P<postdev>(\.post(?P<post>\d+))?(\.dev(?P<dev>\d+))?)?"
)
version_re = re.compile(version_re_str)


def get_python_version(path: str | Path) -> str:
    """
    Get python version string using subprocess from a given path.

    Args:
        path: Path to the Python executable.

    Returns:
        The Python version string.

    Raises:
        InvalidPythonVersion: If the path is not a valid Python executable.
    """
    version_cmd = [
        str(path),
        "-c",
        "import sys; print('.'.join([str(i) for i in sys.version_info[:3]]))",
    ]
    subprocess_kwargs = {
        "env": os.environ.copy(),
        "universal_newlines": True,
        "stdout": subprocess.PIPE,
        "stderr": subprocess.PIPE,
        "shell": False,
    }

    try:
        c = subprocess.Popen(version_cmd, **subprocess_kwargs)
        try:
            out, _ = c.communicate(timeout=5)  # 5 second timeout
        except TypeError:  # For Python versions or mocks that don't support timeout
            out, _ = c.communicate()
    except (SystemExit, KeyboardInterrupt, TimeoutError, subprocess.TimeoutExpired):
        raise InvalidPythonVersion(f"{path} is not a valid python path (timeout)")
    except OSError:
        raise InvalidPythonVersion(f"{path} is not a valid python path")

    if not out:
        raise InvalidPythonVersion(f"{path} is not a valid python path")

    return out.strip()


def parse_python_version(version_str: str) -> dict[str, Any]:
    """
    Parse a Python version string into a dictionary of version components.

    Args:
        version_str: The version string to parse.

    Returns:
        A dictionary containing the parsed version components.

    Raises:
        InvalidPythonVersion: If the version string is not a valid Python version.
    """
    from pipenv.vendor.packaging.version import parse as parse_version

    is_debug = False
    if version_str.endswith("-debug"):
        is_debug = True
        version_str, _, _ = version_str.rpartition("-")

    match = version_re.match(version_str)
    if not match:
        raise InvalidPythonVersion(f"{version_str} is not a python version")

    version_dict = match.groupdict()
    major = int(version_dict.get("major", 0)) if version_dict.get("major") else None
    minor = int(version_dict.get("minor", 0)) if version_dict.get("minor") else None
    patch = int(version_dict.get("patch", 0)) if version_dict.get("patch") else None
    # Initialize release type flags
    is_prerelease = False
    is_devrelease = False
    is_postrelease = False

    try:
        version = parse_version(version_str)
        # Use packaging.version's properties to determine release type
        is_devrelease = version.is_devrelease

        # Check if this is a prerelease
        # A version is a prerelease if:
        # 1. It has a prerelease component (a, b, c, rc) but is not ONLY a dev release
        # 2. For complex versions with both prerelease and dev components, we consider them prereleases
        has_prerelease_component = hasattr(version, "pre") and version.pre is not None
        is_prerelease = has_prerelease_component or (
            version.is_prerelease and not is_devrelease
        )
        # Check for post-release by examining the version string and the version object
        is_postrelease = (hasattr(version, "post") and version.post is not None) or (
            version_dict.get("post") is not None
        )
    except (TypeError, InvalidVersion):
        # If packaging.version can't parse it, try to construct a version string
        # that it can parse
        v_dict = version_dict.copy()
        pre = ""
        if v_dict.get("prerel") and v_dict.get("prerelversion"):
            pre = v_dict.pop("prerel")
            pre = f"{pre}{v_dict.pop('prerelversion')}"
        v_dict["pre"] = pre
        keys = ["major", "minor", "patch", "pre", "postdev", "post", "dev"]
        values = [v_dict.get(val) for val in keys]
        version_str = ".".join([str(v) for v in values if v])
        try:
            version = parse_version(version_str)
            # Update release type flags based on the parsed version
            is_devrelease = version.is_devrelease

            # Check if this is a prerelease
            # A version is a prerelease if:
            # 1. It has a prerelease component (a, b, c, rc) but is not ONLY a dev release
            # 2. For complex versions with both prerelease and dev components, we consider them prereleases
            has_prerelease_component = hasattr(version, "pre") and version.pre is not None
            is_prerelease = has_prerelease_component or (
                version.is_prerelease and not is_devrelease
            )
            # Check for post-release by examining the version string and the version object
            is_postrelease = (hasattr(version, "post") and version.post is not None) or (
                version_dict.get("post") is not None
            )
        except (TypeError, InvalidVersion):
            version = None

    return {
        "major": major,
        "minor": minor,
        "patch": patch,
        "is_postrelease": is_postrelease,
        "is_prerelease": is_prerelease,
        "is_devrelease": is_devrelease,
        "is_debug": is_debug,
        "version": version,
    }


def guess_company(path: str) -> str | None:
    """
    Given a path to python, guess the company who created it.

    Args:
        path: The path to guess about.

    Returns:
        The guessed company name, or "PythonCore" if no match is found.
    """
    from .path_utils import PYTHON_IMPLEMENTATIONS

    non_core_pythons = [impl for impl in PYTHON_IMPLEMENTATIONS if impl != "python"]
    return next(
        iter(impl for impl in non_core_pythons if impl in path.lower()), "PythonCore"
    )


def parse_pyenv_version_order(filename: str = "version") -> list[str]:
    """
    Parse the pyenv version order from the specified file.

    Args:
        filename: The name of the file to parse.

    Returns:
        A list of version strings in the order specified by pyenv.
    """
    from .path_utils import resolve_path

    pyenv_root = os.path.expanduser(
        os.path.expandvars(os.environ.get("PYENV_ROOT", "~/.pyenv"))
    )
    version_order_file = resolve_path(os.path.join(pyenv_root, filename))

    if os.path.exists(version_order_file) and os.path.isfile(version_order_file):
        with open(version_order_file, encoding="utf-8") as fh:
            contents = fh.read()
        version_order = [v for v in contents.splitlines()]
        return version_order

    return []


def parse_asdf_version_order(filename: str = ".tool-versions") -> list[str]:
    """
    Parse the asdf version order from the specified file.

    Args:
        filename: The name of the file to parse.

    Returns:
        A list of version strings in the order specified by asdf.
    """
    from .path_utils import resolve_path

    version_order_file = resolve_path(os.path.join("~", filename))

    if os.path.exists(version_order_file) and os.path.isfile(version_order_file):
        with open(version_order_file, encoding="utf-8") as fh:
            contents = fh.read()
        python_section = next(
            iter(line for line in contents.splitlines() if line.startswith("python")),
            None,
        )
        if python_section:
            # python_key, _, versions
            _, _, versions = python_section.partition(" ")
            if versions:
                return versions.split()

    return []



import datetime
import os
import pathlib
import re
import subprocess
import sys

import invoke
import semver

from pipenv.utils.shell import temp_environ

from .vendoring import _get_git_root, drop_dir

VERSION_FILE = "pipenv/__version__.py"
ROOT = pathlib.Path(".").parent.parent.absolute()
PACKAGE_NAME = "pipenv"


def log(msg):
    print(f"[release] {msg}")


def get_version_file(ctx):
    return _get_git_root(ctx).joinpath(VERSION_FILE)


def find_version(ctx):
    version_file = get_version_file(ctx).read_text()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def get_history_file(ctx):
    return _get_git_root(ctx).joinpath("HISTORY.txt")


def get_dist_dir(ctx):
    return _get_git_root(ctx) / "dist"


def get_build_dir(ctx):
    return _get_git_root(ctx) / "build"


def _render_log(ctx=None, version=None):
    """Totally tap into Towncrier internals to get an in-memory result."""
    if ctx and not version:
        version = find_version(ctx)

    # Use towncrier build --draft with --version flag
    cmd = ["towncrier", "build", "--draft"]
    if version:
        # Add version using the --version flag
        cmd.extend(["--version", version])

    rendered = subprocess.check_output(cmd).decode("utf-8")
    return rendered


release_help = {
    "dry_run": "No-op, simulate what would happen if run for real.",
    "pre": "Build a pre-release version, must be paired with a tag.",
}


@invoke.task(help=release_help)
def release(
    ctx,
    dry_run=False,
    pre=False,
):
    drop_dist_dirs(ctx)
    version = bump_version(
        ctx,
        dry_run=dry_run,
        pre=pre,
    )
    tag_content = _render_log(ctx, version)
    if dry_run:
        # Use the correct version when generating the draft
        ctx.run(f"towncrier build --draft --version {version} > CHANGELOG.draft.md")
        log("would remove: news/*")
        log("would remove: CHANGELOG.draft.md")
        log("would update: pipenv/pipenv.1")
        log(f'Would commit with message: "Release v{version}"')
    else:
        if pre:
            log("generating towncrier draft...")
            # Use the correct version when generating the draft
            ctx.run(f"towncrier build --draft --version {version} > CHANGELOG.draft.md")
            ctx.run(f"git add {get_version_file(ctx).as_posix()}")
        else:
            # Use the correct version when generating the changelog
            ctx.run(f"towncrier build --version {version}")
            ctx.run(f"git add CHANGELOG.md news/ {get_version_file(ctx).as_posix()}")
            log("removing changelog draft if present")
            draft_changelog = pathlib.Path("CHANGELOG.draft.md")
            if draft_changelog.exists():
                draft_changelog.unlink()
        log("generating man files...")
        generate_manual(ctx)
        ctx.run("git add pipenv/pipenv.1")
        ctx.run(f'git commit -m "Release v{version}"')

    tag_content = tag_content.replace('"', '\\"')
    if dry_run or pre:
        log(f"Generated tag content: {tag_content}")
        # draft_rstfile = "CHANGELOG.draft.rst"
        # markdown_path = pathlib.Path(draft_rstfile).with_suffix(".md")
        # generate_markdown(ctx, source_rstfile=draft_rstfile)
        # clean_mdchangelog(ctx, markdown_path.as_posix())
        # log(f"would generate markdown: {markdown_path.read_text()}")
        if not dry_run:
            ctx.run(f'git tag -a v{version} -m "Version v{version}\n\n{tag_content}"')
    else:
        # generate_markdown(ctx)
        # clean_mdchangelog(ctx)
        ctx.run(f'git tag -a v{version} -m "Version v{version}\n\n{tag_content}"')


def drop_dist_dirs(ctx):
    log("Dropping Dist dir...")
    drop_dir(get_dist_dir(ctx))
    log("Dropping build dir...")
    drop_dir(get_build_dir(ctx))


@invoke.task
def build_dists(ctx):
    drop_dist_dirs(ctx)
    py_version = ".".join(str(v) for v in sys.version_info[:2])
    env = {"PIPENV_PYTHON": py_version}
    with ctx.cd(ROOT.as_posix()), temp_environ():
        executable = ctx.run(
            "python -c 'import sys; print(sys.executable)'", hide=True
        ).stdout.strip()
        log(f"Building sdist using {executable} ....")
        os.environ["PIPENV_PYTHON"] = py_version
        ctx.run("pipenv install --dev", env=env)
        ctx.run("pipenv run pip install -e . --upgrade --upgrade-strategy=eager", env=env)
        log(f"Building wheel using python {py_version} ....")
        ctx.run("pipenv run python -m build", env=env)


@invoke.task(build_dists)
def upload_dists(ctx, repo="pypi"):
    dist_pattern = f'{PACKAGE_NAME.replace("-", "[-_]")}-*'
    artifacts = list(ROOT.joinpath("dist").glob(dist_pattern))
    filename_display = "\n".join(f"  {a}" for a in artifacts)
    print(f"[release] Will upload:\n{filename_display}")
    try:
        input("[release] Release ready. ENTER to upload, CTRL-C to abort: ")
    except KeyboardInterrupt:
        print("\nAborted!")
        return

    arg_display = " ".join(f'"{n}"' for n in artifacts)
    ctx.run(f'twine upload --repository="{repo}" {arg_display}')


@invoke.task
def generate_markdown(ctx, source_rstfile=None):
    log("Generating markdown from changelog...")
    if source_rstfile is None:
        source_rstfile = "CHANGELOG.md"
    source_file = pathlib.Path(source_rstfile)
    dest_file = source_file.with_suffix(".md")
    ctx.run(
        f"pandoc {source_file.as_posix()} -f rst -t markdown -o {dest_file.as_posix()}"
    )


@invoke.task
def generate_manual(ctx, commit=False):
    log("Generating manual from reStructuredText source...")
    ctx.run("make man")
    ctx.run("cp docs/_build/man/pipenv.1 pipenv/")
    if commit:
        log("Committing...")
        ctx.run("git add pipenv/pipenv.1")
        ctx.run('git commit -m "Update manual page."')


@invoke.task
def generate_contributing_md(ctx, commit=False):
    log("Generating CONTRIBUTING.md from reStructuredText source...")
    ctx.run("pandoc docs/dev/contributing.rst -f rst -t markdown -o CONTRIBUTING.md")
    if commit:
        log("Committing...")
        ctx.run("git add CONTRIBUTING.md")
        ctx.run('git commit -m "Update CONTRIBUTING.md."')


@invoke.task
def generate_changelog(ctx, commit=False, draft=False):
    log("Generating changelog...")
    # Get the current version to ensure it's used in the changelog
    version = find_version(ctx)
    if draft:
        commit = False
        log("Writing draft to file...")
        ctx.run(f"towncrier build --draft --version {version} > CHANGELOG.draft.md")
    else:
        ctx.run(f"towncrier build --version {version}")
    if commit:
        log("Committing...")
        ctx.run("git add CHANGELOG.md")
        ctx.run("git rm CHANGELOG.draft.md")
        ctx.run('git commit -m "Update changelog."')


@invoke.task
def clean_mdchangelog(ctx, filename=None, content=None):
    changelog = None
    if not content:
        if filename is not None:
            changelog = pathlib.Path(filename)
        else:
            changelog = _get_git_root(ctx) / "CHANGELOG.md"
        content = changelog.read_text()
    content = re.sub(
        r"([^\n]+)\n?\s+\[[\\]+(#\d+)\]\(https://github\.com/pypa/[\w\-]+/issues/\d+\)",
        r"\1 \2",
        content,
        flags=re.MULTILINE,
    )
    if changelog:
        changelog.write_text(content)
    else:
        return content


@invoke.task
def tag_version(ctx, push=False):
    version = find_version(ctx)
    version = semver.VersionInfo.parse(version)
    log(f"Tagging revision: v{version}")
    ctx.run(f"git tag v{version}")
    if push:
        log("Pushing tags...")
        ctx.run("git push origin master")
        ctx.run("git push --tags")


def add_one_day(dt):
    return dt + datetime.timedelta(days=1)


def date_offset(dt, month_offset=0, day_offset=0, truncate=False):
    new_month = (dt.month + month_offset) % 12
    year_offset = new_month // 12
    replace_args = {
        "month": dt.month + month_offset,
        "year": dt.year + year_offset,
    }
    log(
        "Getting updated date from date: {} using month offset: {} and year offset {}".format(
            dt, new_month, replace_args["year"]
        )
    )
    if day_offset:
        dt = dt + datetime.timedelta(days=day_offset)
        log(f"updated date using day offset: {day_offset} => {dt}")
    if truncate:
        log("Truncating...")
        replace_args["day"] = 1
    return dt.replace(**replace_args)


@invoke.task
def bump_version(
    ctx, dry_run=False, pre=False, dev=False, minor=False, major=False, patch=False
):
    version = find_version(ctx)
    current_version = semver.VersionInfo.parse(version)

    # Prompt the user for version change type
    if not minor and not major and not patch:
        while True:
            change_type = input(
                "Enter the version change type (major/minor/patch): "
            ).lower()
            if change_type in ["major", "minor", "patch"]:
                break
            print("Invalid input. Please enter 'major', 'minor', or 'patch'.")

    if minor:
        change_type = "minor"
    if major:
        change_type = "major"
    if patch:
        change_type = "patch"

    # Bump the version based on the user input
    if change_type == "major":
        new_version = current_version.bump_major()
    elif change_type == "minor":
        new_version = current_version.bump_minor()
    else:
        new_version = current_version.bump_patch()
    # Pre-release handling code
    if pre:
        new_version = new_version.bump_prerelease()
    if dev:
        new_version = new_version.bump_prerelease("dev")

    # Update the version file
    log(f"Found current version: {version}")
    log(f"Updating version to {new_version}")

    if dry_run:
        sys.exit(0)
    else:
        log(f"Updating to: {new_version}")
        version_file = get_version_file(ctx)
        file_contents = version_file.read_text()
        version_file.write_text(file_contents.replace(version, str(new_version)))
        ctx.run(f"git add {version_file.as_posix()}")
        log("Committing...")
        ctx.run(f'git commit -s -m "Bumped version to {new_version}."')
    return str(new_version)



import os
from pathlib import Path

import pytest


@pytest.mark.basic
@pytest.mark.install
def test_install_github_vcs(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        c = p.pipenv("install git+https://github.com/reagento/adaptix.git@2.16")
        assert not c.returncode
        assert "dataclass-factory" in p.pipfile["packages"]


@pytest.mark.basic
@pytest.mark.install
@pytest.mark.parametrize("use_credentials", [True, False])
def test_install_github_vcs_with_credentials(pipenv_instance_pypi, use_credentials):
    with pipenv_instance_pypi() as p:
        # Set environment variables
        os.environ['GIT_REPO'] = 'github.com/reagento/adaptix.git'
        if use_credentials:
            os.environ['GIT_USERNAME'] = 'git'  # Use 'git' as a dummy username
            os.environ['GIT_PASSWORD'] = ''  # Empty password for public repos
            url = "git+https://${GIT_USERNAME}:${GIT_PASSWORD}@${GIT_REPO}@2.16"
        else:
            url = "git+https://${GIT_REPO}@2.16"
        if os.name == 'nt':
            c = p.pipenv(f"install {url} -v")
        else:
            c = p.pipenv(f"install '{url}' -v")
        assert c.returncode == 0, f"Install failed with error: {c.stderr}"

        assert "dataclass-factory" in p.pipfile["packages"]

        # Check if the URL in the lockfile still contains the environment variables
        lockfile_content = p.lockfile
        assert "${GIT_REPO}" in lockfile_content['default']['dataclass-factory']['git']
        if use_credentials:
            assert "${GIT_USERNAME}" in lockfile_content['default']['dataclass-factory']['git']
            assert "${GIT_PASSWORD}" in lockfile_content['default']['dataclass-factory']['git']

        # Verify that the package is installed and usable
        c = p.pipenv("run python -c \"import dataclass_factory\"")
        assert c.returncode == 0, f"Failed to import library: {c.stderr}"


@pytest.mark.vcs
@pytest.mark.urls
@pytest.mark.install
@pytest.mark.needs_internet
def test_install_vcs_ref_by_commit_hash(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        c = p.pipenv("install -e git+https://github.com/benjaminp/six.git@5efb522b0647f7467248273ec1b893d06b984a59#egg=six")
        assert c.returncode == 0
        assert "six" in p.pipfile["packages"]
        assert "six" in p.lockfile["default"]
        assert (
            p.lockfile["default"]["six"]["ref"]
            == "5efb522b0647f7467248273ec1b893d06b984a59"
        )
        pipfile = Path(p.pipfile_path)
        new_content = pipfile.read_text().replace("5efb522b0647f7467248273ec1b893d06b984a59", "15e31431af97e5e64b80af0a3f598d382bcdd49a")
        pipfile.write_text(new_content)
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert (
            p.lockfile["default"]["six"]["ref"]
            == "15e31431af97e5e64b80af0a3f598d382bcdd49a"
        )
        assert "six" in p.pipfile["packages"]
        assert "six" in p.lockfile["default"]


@pytest.mark.vcs
@pytest.mark.dev
@pytest.mark.install
def test_vcs_dev_package_install(pipenv_instance_pypi):
    """Ensure VCS packages can be properly installed into dev-packages via --dev flag with existing Pipfile."""
    with pipenv_instance_pypi() as p:
        # Create a Pipfile with some existing packages
        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
six = "*"

[dev-packages]
pytest-xdist = {git = "https://github.com/pytest-dev/pytest-xdist.git", ref = "v3.6.1"}
            """.strip()
            f.write(contents)

        # Install a VCS package with --dev flag
        c = p.pipenv("install --dev -v")
        assert c.returncode == 0

        # Verify package appears in dev-packages in Pipfile
        assert "pytest-xdist" in p.pipfile["dev-packages"]
        assert p.pipfile["dev-packages"]["pytest-xdist"]["git"] == "https://github.com/pytest-dev/pytest-xdist.git"
        assert p.pipfile["dev-packages"]["pytest-xdist"]["ref"] == "v3.6.1"

        # Verify package appears in develop section of lockfile
        assert "pytest-xdist" in p.lockfile["develop"]
        assert p.lockfile["develop"]["pytest-xdist"]["git"] == "https://github.com/pytest-dev/pytest-xdist.git"
        assert p.lockfile["develop"]["pytest-xdist"]["ref"] == "4dd2978031eaf7017c84a1cc77667379a2b11c64"

        # Verify the package is importable
        c = p.pipenv('run python -c "import xdist"')
        assert c.returncode == 0

        # Test that dependencies were also installed correctly
        c = p.pipenv('run python -c "import execnet"')  # pytest-xdist depends on execnet
        assert c.returncode == 0

        # Verify no duplicate entries in default packages
        assert "pytest-xdist" not in p.pipfile.get("packages", {})
        assert "pytest-xdist" not in p.lockfile.get("default", {})



import os
from pathlib import Path

import pytest
from flaky import flaky

from pipenv.utils.shell import temp_environ


@pytest.mark.lock
@pytest.mark.requirements
def test_lock_handle_eggs(pipenv_instance_private_pypi):
    """Ensure locking works with packages providing egg formats."""
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            f.write(
                f"""
[[source]]
url = "{p.index_url}"
verify_ssl = false
name = "testindex"

[packages]
RandomWords = "*"
            """
            )
        c = p.pipenv("lock --verbose")
        assert c.returncode == 0
        assert "randomwords" in p.lockfile["default"]
        assert p.lockfile["default"]["randomwords"]["version"] == "==0.2.1"


@pytest.mark.lock
@pytest.mark.requirements
def test_lock_gathers_pyproject_dependencies(pipenv_instance_pypi):
    """Ensure that running `pipenv install` doesn't install dev packages"""
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
pipenvtest = { editable = true, path = "." }
            """.strip()
            f.write(contents)

        # Write the pyproject.toml
        pyproject_toml_path = os.path.join(
            os.path.dirname(p.pipfile_path), "pyproject.toml"
        )
        with open(pyproject_toml_path, "w") as f:
            contents = """
[build-system]
requires = ["setuptools"]
build-backend = "setuptools.build_meta"

[project]
name = "pipenvtest"
version = "0.0.1"
requires-python = ">=3.8"
dependencies = [
    "six"
]
            """.strip()
            f.write(contents)
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "six" in p.lockfile["default"]


@pytest.mark.lock
@pytest.mark.requirements
def test_lock_requirements_file(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
urllib3 = "==1.23"
[dev-packages]
colorama = "==0.3.9"
            """.strip()
            f.write(contents)

        req_list = ("urllib3==1.23",)

        dev_req_list = ("colorama==0.3.9",)

        c = p.pipenv("lock")
        assert c.returncode == 0

        default = p.pipenv("requirements")
        assert default.returncode == 0
        dev = p.pipenv("requirements --dev-only")

        for req in req_list:
            assert req in default.stdout

        for req in dev_req_list:
            assert req in dev.stdout


@pytest.mark.lock
def test_lock_includes_hashes_for_all_platforms(pipenv_instance_private_pypi):
    """Locking should include hashes for *all* platforms, not just the
    platform we're running lock on."""

    # releases = pytest_pypi.app.packages['yarl'].releases

    releases = {
        "yarl-1.3.0-cp35-cp35m-manylinux1_x86_64.whl": "3890ab952d508523ef4881457c4099056546593fa05e93da84c7250516e632eb",
        "yarl-1.3.0-cp35-cp35m-win_amd64.whl": "b25de84a8c20540531526dfbb0e2d2b648c13fd5dd126728c496d7c3fea33310",
        "yarl-1.3.0-cp36-cp36m-manylinux1_x86_64.whl": "5badb97dd0abf26623a9982cd448ff12cb39b8e4c94032ccdedf22ce01a64842",
        "yarl-1.3.0-cp36-cp36m-win_amd64.whl": "c6e341f5a6562af74ba55205dbd56d248daf1b5748ec48a0200ba227bb9e33f4",
        "yarl-1.3.0-cp37-cp37m-win_amd64.whl": "73f447d11b530d860ca1e6b582f947688286ad16ca42256413083d13f260b7a0",
        "yarl-1.3.0.tar.gz": "024ecdc12bc02b321bc66b41327f930d1c2c543fa9a561b39861da9388ba7aa9",
    }

    def get_hash(release_name):
        # Convert a specific filename to a hash like what would show up in a Pipfile.lock.
        # For example:
        # 'yarl-1.3.0-cp35-cp35m-manylinux1_x86_64.whl' -> 'sha256:3890ab952d508523ef4881457c4099056546593fa05e93da84c7250516e632eb'
        return f"sha256:{releases[release_name]}"

    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = false
name = "testindex"

[packages]
yarl = "==1.3.0"
            """.strip()
            f.write(contents)

        c = p.pipenv("lock")
        assert c.returncode == 0

        lock = p.lockfile
        assert "yarl" in lock["default"]
        assert set(lock["default"]["yarl"]["hashes"]) == {
            get_hash("yarl-1.3.0-cp35-cp35m-manylinux1_x86_64.whl"),
            get_hash("yarl-1.3.0-cp35-cp35m-win_amd64.whl"),
            get_hash("yarl-1.3.0-cp36-cp36m-manylinux1_x86_64.whl"),
            get_hash("yarl-1.3.0-cp36-cp36m-win_amd64.whl"),
            get_hash("yarl-1.3.0-cp37-cp37m-win_amd64.whl"),
            get_hash("yarl-1.3.0.tar.gz"),
        }


@pytest.mark.lock
def test_resolve_skip_unmatched_requirements(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        p._pipfile.add("missing-package", {"markers": "os_name=='FakeOS'"})
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert (
            'Could not find a matching version of missing-package; os_name == "FakeOS"'
            in c.stderr
        )


@pytest.mark.lock
@pytest.mark.complex
@pytest.mark.needs_internet
def test_complex_lock_with_vcs_deps(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        requests_uri = p._pipfile.get_fixture_path("git/requests").as_uri()
        dateutil_uri = p._pipfile.get_fixture_path("git/dateutil").as_uri()
        with open(p.pipfile_path, "w") as f:
            contents = (
                """
[packages]
click = "==6.7"

[dev-packages]
requests = {git = "%s"}
            """.strip()
                % requests_uri
            )
            f.write(contents)

        c = p.pipenv("install")
        assert c.returncode == 0
        lock = p.lockfile
        assert "requests" in lock["develop"]
        assert "click" in lock["default"]

        c = p.pipenv(f"run pip install -e git+{dateutil_uri}#egg=python_dateutil")
        assert c.returncode == 0

        lock = p.lockfile
        assert "requests" in lock["develop"]
        assert "click" in lock["default"]
        assert "python_dateutil" not in lock["default"]
        assert "python_dateutil" not in lock["develop"]


@pytest.mark.lock
@pytest.mark.requirements
def test_lock_with_prereleases(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
sqlalchemy = "==1.2.0b3"

[pipenv]
allow_prereleases = true
            """.strip()
            f.write(contents)

        c = p.pipenv("lock")
        assert c.returncode == 0
        assert p.lockfile["default"]["sqlalchemy"]["version"] == "==1.2.0b3"


@pytest.mark.lock
@pytest.mark.maya
@pytest.mark.complex
@pytest.mark.needs_internet
@flaky
def test_complex_deps_lock_and_install_properly(pipenv_instance_pypi):
    # This uses the real PyPI because Maya has too many dependencies...
    with pipenv_instance_pypi() as p, open(p.pipfile_path, "w") as f:
        contents = """
[packages]
maya = "*"
            """.strip()
        f.write(contents)

        c = p.pipenv("lock --verbose")
        assert c.returncode == 0

        c = p.pipenv("install")
        assert c.returncode == 0


@pytest.mark.lock
@pytest.mark.extras
def test_lock_extras_without_install(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
requests = {version = "*", extras = ["socks"]}
            """.strip()
            f.write(contents)

        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]
        assert "pysocks" in p.lockfile["default"]
        assert "markers" in p.lockfile["default"]["pysocks"]

        c = p.pipenv("lock")
        assert c.returncode == 0
        c = p.pipenv("requirements")
        assert c.returncode == 0
        assert "extra == 'socks'" not in c.stdout.strip()


@pytest.mark.lock
@pytest.mark.index
@pytest.mark.install  # private indexes need to be uncached for resolution
@pytest.mark.requirements
@pytest.mark.needs_internet
def test_private_index_lock_requirements(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[[source]]
url = "https://test.pypi.org/simple"
verify_ssl = true
name = "testpypi"

[packages]
pipenv-test-private-package = {version = "*", index = "testpypi"}
            """.strip()
            f.write(contents)
        c = p.pipenv("lock")
        assert c.returncode == 0


@pytest.mark.lock
@pytest.mark.index
@pytest.mark.install  # private indexes need to be uncached for resolution
@pytest.mark.requirements
@pytest.mark.needs_internet
def test_private_index_lock_requirements_for_not_canonical_package(
    pipenv_instance_private_pypi,
):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[[source]]
url = "https://test.pypi.org/simple"
verify_ssl = true
name = "testpypi"

[packages]
pipenv_test_private_package = {version = "*", index = "testpypi"}
            """.strip()
            f.write(contents)
        c = p.pipenv("lock")
        assert c.returncode == 0


@pytest.mark.index
@pytest.mark.install
def test_lock_updated_source(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "{url}/${{MY_ENV_VAR}}"
name = "localpypi"
verify_ssl = false

[packages]
requests = "==2.14.0"
            """.strip().format(
                url=p.pypi
            )
            f.write(contents)

        with temp_environ():
            os.environ["MY_ENV_VAR"] = "simple"
            c = p.pipenv("lock")
            assert c.returncode == 0
            assert "requests" in p.lockfile["default"]

        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "{url}/simple"
name = "localpypi"
verify_ssl = false

[packages]
requests = "==2.14.0"
            """.strip().format(
                url=p.pypi
            )
            f.write(contents)

        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]


@pytest.mark.vcs
@pytest.mark.lock
@pytest.mark.needs_internet
def test_lock_editable_vcs_without_install(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        requests_uri = p._pipfile.get_fixture_path("git/six").as_uri()
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[packages]
six = {git = "%s", editable = true}
            """.strip()
                % requests_uri
            )
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "six" in p.lockfile["default"]


@pytest.mark.vcs
@pytest.mark.lock
@pytest.mark.needs_internet
def test_lock_editable_vcs_with_ref_in_git(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        requests_uri = p._pipfile.get_fixture_path("git/requests").as_uri()
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[packages]
requests = {git = "%s@883caaf", editable = true}
            """.strip()
                % requests_uri
            )
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert requests_uri in p.lockfile["default"]["requests"]["git"]
        assert (
            p.lockfile["default"]["requests"]["ref"]
            == "883caaf145fbe93bd0d208a6b864de9146087312"
        )


@pytest.mark.vcs
@pytest.mark.lock
@pytest.mark.extras
@pytest.mark.needs_internet
def test_lock_editable_vcs_with_extras_without_install(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        requests_uri = p._pipfile.get_fixture_path("git/requests").as_uri()
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[packages]
requests = {git = "%s", editable = true, extras = ["socks"]}
            """.strip()
                % requests_uri
            )
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]
        assert "idna" in p.lockfile["default"]
        assert "certifi" in p.lockfile["default"]
        assert "socks" in p.lockfile["default"]["requests"]["extras"]
        assert "version" not in p.lockfile["default"]["requests"]


@pytest.mark.vcs
@pytest.mark.lock
@pytest.mark.needs_internet
def test_lock_editable_vcs_with_markers_without_install(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        requests_uri = p._pipfile.get_fixture_path("git/requests").as_uri()
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[packages]
requests = {git = "%s", editable = true, markers = "python_version >= '2.6'"}
            """.strip()
                % requests_uri
            )
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]
        assert "idna" in p.lockfile["default"]
        assert "certifi" in p.lockfile["default"]
        assert c.returncode == 0


@pytest.mark.lock
@pytest.mark.install
def test_lockfile_corrupted(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        with open(p.lockfile_path, "w") as f:
            f.write("{corrupted}")
        c = p.pipenv("install")
        assert c.returncode != 0
        assert "Pipfile.lock is corrupt" in c.stderr


@pytest.mark.lock
def test_lockfile_with_empty_dict(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        with open(p.lockfile_path, "w") as f:
            f.write("{}")
        c = p.pipenv("install")
        assert c.returncode == 0
        assert p.lockfile["_meta"]


@pytest.mark.vcs
@pytest.mark.lock
def test_vcs_lock_respects_top_level_pins(pipenv_instance_private_pypi):
    """Test that locking VCS dependencies respects top level packages pinned in Pipfiles"""

    with pipenv_instance_private_pypi() as p:
        requests_uri = p._pipfile.get_fixture_path("git/requests").as_uri()
        p._pipfile.add(
            "requests", {"editable": True, "git": f"{requests_uri}", "ref": "v2.18.4"}
        )
        p._pipfile.add("urllib3", "==1.21.1")
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]
        assert "git" in p.lockfile["default"]["requests"]
        assert "urllib3" in p.lockfile["default"]
        assert p.lockfile["default"]["urllib3"]["version"] == "==1.21.1"


@pytest.mark.lock
def test_lock_after_update_source_name(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = true
name = "test"

[packages]
six = "*"
        """.strip()
        with open(p.pipfile_path, "w") as f:
            f.write(contents)
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert p.lockfile["default"]["six"]["index"] == "test"
        with open(p.pipfile_path, "w") as f:
            f.write(contents.replace('name = "test"', 'name = "custom"'))
        c = p.pipenv("lock --clear")
        assert c.returncode == 0
        assert "index" in p.lockfile["default"]["six"]
        assert p.lockfile["default"]["six"]["index"] == "custom", Path(
            p.lockfile_path
        ).read_text()


@pytest.mark.lock
def test_lock_nested_direct_url(pipenv_instance_private_pypi):
    """
    The dependency 'test_package' has a declared dependency on
    a PEP508 style VCS URL. This ensures that we capture the dependency
    here along with its own dependencies.
    """
    with pipenv_instance_private_pypi(pipfile=False) as p:
        contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = true
name = "local"

[packages]
test_package = "*"
                """.strip()
        with open(p.pipfile_path, "w") as f:
            f.write(contents)
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "vistir" in p.lockfile["default"]
        assert "colorama" in p.lockfile["default"]
        assert "six" in p.lockfile["default"]


@pytest.mark.lock
@pytest.mark.needs_internet
def test_lock_nested_vcs_direct_url(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        p._pipfile.add(
            "pep508_package",
            {
                "git": "https://github.com/techalchemy/test-project.git",
                "editable": True,
                "ref": "master",
                "subdirectory": "parent_folder/pep508-package",
            },
        )
        c = p.pipenv("lock -v")
        assert c.returncode == 0
        assert "git" in p.lockfile["default"]["pep508-package"]
        assert "sibling-package" in p.lockfile["default"]
        assert "git" in p.lockfile["default"]["sibling-package"]
        assert "subdirectory" in p.lockfile["default"]["sibling-package"]
        assert (
            p.lockfile["default"]["sibling-package"]["subdirectory"]
            == "parent_folder/sibling_package"
        )
        assert "version" not in p.lockfile["default"]["sibling-package"]


@pytest.mark.lock
@pytest.mark.install
def test_lock_package_with_compatible_release_specifier(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        c = p.pipenv("install six~=1.11")
        assert c.returncode == 0
        assert "six" in p.pipfile["packages"]
        assert p.pipfile["packages"]["six"] == "~=1.11"
        assert "six" in p.lockfile["default"]
        assert "version" in p.lockfile["default"]["six"]
        assert p.lockfile["default"]["six"]["version"] == "==1.12.0"


@pytest.mark.lock
@pytest.mark.install
def test_default_lock_overwrite_dev_lock(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        c = p.pipenv("install click==6.7")
        assert c.returncode == 0
        c = p.pipenv("install -d flask")
        assert c.returncode == 0
        assert p.lockfile["default"]["click"]["version"] == "==6.7"
        assert p.lockfile["develop"]["click"]["version"] == "==6.7"


@flaky
@pytest.mark.lock
@pytest.mark.install
@pytest.mark.needs_internet
def test_pipenv_respects_package_index_restrictions(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[[source]]
url = "{url}"
verify_ssl = true
name = "local"

[packages]
requests = {requirement}
                """.strip().format(
                url=p.index_url, requirement='{version="*", index="local"}'
            )
            f.write(contents)

        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]
        assert "idna" in p.lockfile["default"]
        assert "certifi" in p.lockfile["default"]
        assert "urllib3" in p.lockfile["default"]
        assert "chardet" in p.lockfile["default"]
        # this is the newest version we have in our private pypi (but pypi.org has 2.27.1 at present)
        expected_result = {
            "hashes": [
                "sha256:63b52e3c866428a224f97cab011de738c36aec0185aa91cfacd418b5d58911d1",
                "sha256:ec22d826a36ed72a7358ff3fe56cbd4ba69dd7a6718ffd450ff0e9df7a47ce6a",
            ],
            "index": "local",
            "version": "==2.19.1",
        }
        assert p.lockfile["default"]["requests"] == expected_result


@pytest.mark.dev
@pytest.mark.lock
@pytest.mark.install
def test_dev_lock_use_default_packages_as_constraint(pipenv_instance_private_pypi):
    # See https://github.com/pypa/pipenv/issues/4371
    # See https://github.com/pypa/pipenv/issues/2987
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = false
name = "testindex"

[packages]
requests = "<=2.14.0"

[dev-packages]
requests = "*"
                """.strip()
            f.write(contents)

        c = p.pipenv("lock --dev")
        assert c.returncode == 0
        assert "requests" in p.lockfile["default"]
        assert p.lockfile["default"]["requests"]["version"] == "==2.14.0"
        assert "requests" in p.lockfile["develop"]
        assert p.lockfile["develop"]["requests"]["version"] == "==2.14.0"

        # requests 2.14.0 doesn't require these packages
        assert "idna" not in p.lockfile["develop"]
        assert "certifi" not in p.lockfile["develop"]
        assert "urllib3" not in p.lockfile["develop"]
        assert "chardet" not in p.lockfile["develop"]


@pytest.mark.lock
def test_lock_specific_named_category(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi(pipfile=False) as p:
        contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = true
name = "test"

[packages]
requests = "*"

[prereq]
six = "*"
        """.strip()
        with open(p.pipfile_path, "w") as f:
            f.write(contents)
        c = p.pipenv("lock --categories prereq")
        assert c.returncode == 0
        assert p.lockfile["prereq"]["six"]["index"] == "test"
        assert p.lockfile["default"] == {}
        c = p.pipenv("lock --categories packages")
        assert c.returncode == 0
        assert p.lockfile["prereq"]["six"]["index"] == "test"
        assert p.lockfile["default"]["requests"]["index"] == "test"


def test_pinned_pipfile_no_null_markers_when_extras(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
dataclasses-json = {extras = ["dev"], version = "==0.5.7"}
            """.strip()
            f.write(contents)
        c = p.pipenv("lock")
        assert c.returncode == 0
        assert "dataclasses-json" in p.pipfile["packages"]
        assert "dataclasses-json" in p.lockfile["default"]
        assert p.lockfile["default"]["dataclasses-json"].get("markers", "") is not None


@pytest.mark.index
@pytest.mark.install  # private indexes need to be uncached for resolution
@pytest.mark.skip_lock
@pytest.mark.needs_internet
def test_private_index_skip_lock(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[[source]]
url = "https://test.pypi.org/simple"
verify_ssl = true
name = "testpypi"

[packages]
pipenv-test-private-package = {version = "*", index = "testpypi"}

[pipenv]
install_search_all_sources = true
            """.strip()
            f.write(contents)
        c = p.pipenv("install --skip-lock")
        assert c.returncode == 0



import os
import tempfile
from pathlib import Path
from unittest import mock

import pytest

from pipenv.patched.pip._internal.operations.prepare import File
from pipenv.project import Project
from pipenv.utils.requirements import import_requirements


@pytest.mark.cli
@pytest.mark.deploy
@pytest.mark.system
@mock.patch(
    "pipenv.utils.dependencies.unpack_url",
    mock.MagicMock(return_value=File("/some/path/to/project", content_type=None)),
)
@mock.patch("pipenv.utils.dependencies.find_package_name_from_directory")
def test_auth_with_pw_redacted(
    mock_find_package_name_from_directory, pipenv_instance_pypi
):
    mock_find_package_name_from_directory.return_value = "myproject"
    with pipenv_instance_pypi() as p:
        p.pipenv("shell")
        project = Project()
        requirements_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
        requirements_file.write(
            """git+https://${AUTH_USER}:mypw1@github.com/user/myproject.git@main#egg=myproject"""
        )
        requirements_file.close()
        import_requirements(project, r=requirements_file.name)
        os.unlink(requirements_file.name)
        assert p.pipfile["packages"]["myproject"] == {'git': 'https://${AUTH_USER}:****@github.com/user/myproject.git', 'ref': 'main'}


@pytest.mark.cli
@pytest.mark.deploy
@pytest.mark.system
@mock.patch(
    "pipenv.utils.dependencies.unpack_url",
    mock.MagicMock(return_value=File("/some/path/to/project", content_type=None)),
)
@mock.patch("pipenv.utils.dependencies.find_package_name_from_directory")
def test_auth_with_username_redacted(
    mock_find_package_name_from_directory, pipenv_instance_pypi
):
    mock_find_package_name_from_directory.return_value = "myproject"
    with pipenv_instance_pypi() as p:
        p.pipenv("shell")
        project = Project()
        requirements_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
        requirements_file.write(
            """git+https://username@github.com/user/myproject.git@main#egg=myproject"""
        )
        requirements_file.close()
        import_requirements(project, r=requirements_file.name)
        os.unlink(requirements_file.name)
        assert p.pipfile["packages"]["myproject"] == {'git': 'https://****@github.com/user/myproject.git', 'ref': 'main'}



@pytest.mark.cli
@pytest.mark.deploy
@pytest.mark.system
@mock.patch(
    "pipenv.utils.dependencies.unpack_url",
    mock.MagicMock(return_value=File("/some/path/to/project", content_type=None)),
)
@mock.patch("pipenv.utils.dependencies.find_package_name_from_directory")
def test_auth_with_pw_are_variables_passed_to_pipfile(
    mock_find_package_name_from_directory, pipenv_instance_pypi
):
    mock_find_package_name_from_directory.return_value = "myproject"
    with pipenv_instance_pypi() as p:
        p.pipenv("shell")
        project = Project()
        requirements_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
        requirements_file.write(
            """git+https://${AUTH_USER}:${AUTH_PW}@github.com/user/myproject.git@main#egg=myproject"""
        )
        requirements_file.close()
        import_requirements(project, r=requirements_file.name)
        os.unlink(requirements_file.name)
        expected = {'git': 'https://${AUTH_USER}:${AUTH_PW}@github.com/user/myproject.git', 'ref': 'main'}
        assert p.pipfile["packages"]["myproject"] == expected


@pytest.mark.cli
@pytest.mark.deploy
@pytest.mark.system
@mock.patch(
    "pipenv.utils.dependencies.unpack_url",
    mock.MagicMock(return_value=File("/some/path/to/project", content_type=None)),
)
@mock.patch("pipenv.utils.dependencies.find_package_name_from_directory")
def test_auth_with_only_username_variable_passed_to_pipfile(
    mock_find_package_name_from_directory, pipenv_instance_pypi
):
    mock_find_package_name_from_directory.return_value = "myproject"
    with pipenv_instance_pypi() as p:
        p.pipenv("shell")
        project = Project()
        requirements_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
        requirements_file.write(
            """git+https://${AUTH_USER}@github.com/user/myproject.git@main#egg=myproject"""
        )
        requirements_file.close()
        import_requirements(project, r=requirements_file.name)
        os.unlink(requirements_file.name)
        assert p.pipfile["packages"]["myproject"] == {'git': 'https://${AUTH_USER}@github.com/user/myproject.git', 'ref': 'main'}



@pytest.mark.integration
def test_import_requirements_with_path_object(pipenv_instance_pypi):
    """Test that import_requirements can handle Path objects correctly."""
    # Create a temporary requirements.txt file
    with pipenv_instance_pypi() as p:
        req_file_path = Path(p.path) / "requirements.txt"
        with open(req_file_path, "w") as f:
            f.write("requests\n")

        try:
            # Create a Project instance
            project = Project()

            # Call import_requirements with a Path object
            import_requirements(project, r=req_file_path)

            # Verify that the package was added to the Pipfile
            with open("Pipfile") as f:
                pipfile_content = f.read()
                assert "requests" in pipfile_content
        finally:
            # Clean up
            if os.path.exists(req_file_path):
                os.unlink(req_file_path)
            if os.path.exists("Pipfile"):
                os.unlink("Pipfile")


@pytest.mark.integration
def test_import_requirements_preserves_markers(pipenv_instance_pypi):
    """Test that import_requirements preserves environment markers like sys_platform.

    Fixes: https://github.com/pypa/pipenv/issues/6101
    """
    with pipenv_instance_pypi() as p:
        req_file_path = Path(p.path) / "requirements.txt"
        # Create requirements with environment markers
        with open(req_file_path, "w") as f:
            f.write("six; sys_platform == 'win32'\n")

        try:
            # Create a Project instance
            project = Project()

            # Call import_requirements
            import_requirements(project, r=req_file_path)

            # Verify that the package was added with markers
            assert "six" in p.pipfile["packages"]
            pkg_entry = p.pipfile["packages"]["six"]
            # The entry should be a dict with markers
            assert isinstance(pkg_entry, dict)
            assert "markers" in pkg_entry
            assert "sys_platform" in pkg_entry["markers"]
            assert "win32" in pkg_entry["markers"]
        finally:
            # Clean up
            if os.path.exists(req_file_path):
                os.unlink(req_file_path)
            if os.path.exists("Pipfile"):
                os.unlink("Pipfile")



import os
import sys

import pytest
from flaky import flaky

from pipenv.project import Project
from pipenv.utils.shell import temp_environ


@pytest.mark.markers
def test_package_environment_markers(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[[source]]
url = "{}"
verify_ssl = false
name = "pypi"

[packages]
dataclass-factory = {}

[dev-packages]
            """.format(
                p.index_url,
                '{version = "*", os_name = "== \'splashwear\'", index="pypi"}',
            ).strip()
            f.write(contents)

        c = p.pipenv("install -v")
        assert c.returncode == 0
        assert "markers" in p.lockfile["default"]["dataclass-factory"], p.lockfile["default"]
        assert (
            p.lockfile["default"]["dataclass-factory"]["markers"] == "python_version >= '3.6' and os_name == 'splashwear'"
        ), p.lockfile["default"]["dataclass-factory"]["markers"]
        c = p.pipenv('run python -c "import dataclass_factory;"')
        assert c.returncode == 1  # dataclass-factory is not installed due to the marker


@flaky
@pytest.mark.markers
def test_platform_python_implementation_marker(pipenv_instance_private_pypi):
    """Markers should be converted during locking to help users who input this
    incorrectly.
    """
    with pipenv_instance_private_pypi() as p:
        c = p.pipenv("install depends-on-marked-package")
        assert c.returncode == 0

        # depends-on-marked-package has an install_requires of
        # 'pytz; platform_python_implementation=="CPython"'
        # Verify that that marker shows up in our lockfile unaltered.
        assert "pytz" in p.lockfile["default"]
        assert (
            p.lockfile["default"]["pytz"].get("markers")
            == "platform_python_implementation == 'CPython'"
        )


@flaky
@pytest.mark.alt
@pytest.mark.markers
@pytest.mark.install
def test_specific_package_environment_markers(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
six = {version = "*", os_name = "== 'splashwear'"}
            """.strip()
            f.write(contents)

        c = p.pipenv("install")
        assert c.returncode == 0
        assert "markers" in p.lockfile["default"]["six"]

        c = p.pipenv('run python -c "import six;"')
        assert c.returncode == 1


@flaky
@pytest.mark.markers
def test_top_level_overrides_environment_markers(pipenv_instance_pypi):
    """Top-level environment markers should take precedence."""
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
apscheduler = "*"
funcsigs = {version = "*", os_name = "== 'splashwear'"}
            """.strip()
            f.write(contents)

        c = p.pipenv("install")
        assert c.returncode == 0
        assert "markers" in p.lockfile["default"]["funcsigs"], p.lockfile["default"][
            "funcsigs"
        ]
        assert (
            p.lockfile["default"]["funcsigs"]["markers"] == "os_name == 'splashwear'"
        ), p.lockfile["default"]["funcsigs"]


@flaky
@pytest.mark.markers
@pytest.mark.install
def test_global_overrides_environment_markers(pipenv_instance_private_pypi):
    """Empty (unconditional) dependency should take precedence.
    If a dependency is specified without environment markers, it should
    override dependencies with environment markers. In this example,
    APScheduler requires funcsigs only on Python 2, but since funcsigs is
    also specified as an unconditional dep, its markers should be empty.
    """
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
url = "{p.index_url}"
verify_ssl = false
name = "testindex"

[packages]
apscheduler = "*"
funcsigs = "*"
            """.strip()
            f.write(contents)

        c = p.pipenv("install")
        assert c.returncode == 0

        assert p.lockfile["default"]["funcsigs"].get("markers", "") == ""


@flaky
@pytest.mark.markers
@pytest.mark.complex
@pytest.mark.skipif(
    sys.version_info[:2] == (3, 8), reason="Test package that gets installed is different on 3.8"
)
def test_resolver_unique_markers(pipenv_instance_pypi):
    """Test that markers are properly cleaned and not duplicated when resolving
    dependencies. Use vcrpy as an example package that pulls in dependencies
    with Python version markers.

    This test verifies that even if a package ends up with duplicate markers like:
    'yarl; python_version >= "3.9"; python_version >= "3.9"'

    The resolver will clean and deduplicate them appropriately.
    """
    with pipenv_instance_pypi() as p:
        c = p.pipenv("install vcrpy==2.0.1")
        assert c.returncode == 0
        assert "yarl" in p.lockfile["default"]
        yarl = p.lockfile["default"]["yarl"]
        assert "markers" in yarl
        # Check for a valid Python version marker
        # yarl >=1.16.0 (Oct 2024) requires Python >=3.9
        assert yarl["markers"] == "python_version >= '3.9'"


@pytest.mark.markers
@pytest.mark.install
@pytest.mark.needs_internet
@pytest.mark.skipif(
    os.name == "nt",
    reason="This dependency is not available on Windows",
)
def test_install_package_with_invalid_python_version_specifier(pipenv_instance_pypi):
    """Test that installing a package with an invalid Python version specifier
    doesn't raise a KeyError. This test verifies the fix for issue #6370.
    """
    with pipenv_instance_pypi() as p:
        # typedb-driver 3.0.5 has an invalid Python version specifier that was causing a KeyError
        c = p.pipenv("install typedb-driver==3.0.5")
        assert c.returncode == 0
        assert "typedb-driver" in p.pipfile["packages"]
        assert "typedb-driver" in p.lockfile["default"]


@flaky
@pytest.mark.project
@pytest.mark.needs_internet
def test_environment_variable_value_does_not_change_hash(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p, temp_environ():
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[[source]]
url = 'https://${PYPI_USERNAME}:${PYPI_PASSWORD}@pypi.org/simple'
verify_ssl = true
name = 'pypi'

[packages]
six = "*"
"""
            )
        project = Project()

        os.environ["PYPI_USERNAME"] = "whatever"
        os.environ["PYPI_PASSWORD"] = "pass"
        assert project.get_lockfile_hash() is None

        c = p.pipenv("install")
        assert c.returncode == 0
        lock_hash = project.get_lockfile_hash()
        assert lock_hash is not None
        assert lock_hash == project.calculate_pipfile_hash()

        assert c.returncode == 0
        assert project.get_lockfile_hash() == project.calculate_pipfile_hash()

        os.environ["PYPI_PASSWORD"] = "pass2"
        assert project.get_lockfile_hash() == project.calculate_pipfile_hash()

        with open(p.pipfile_path, "a") as f:
            f.write('requests = "==2.14.0"\n')
        assert project.get_lockfile_hash() != project.calculate_pipfile_hash()



# Copied from pip's vendoring process
# see https://github.com/pypa/pip/blob/95bcf8c5f6394298035a7332c441868f3b0169f4/tasks/__init__.py
from pathlib import Path

import invoke

from . import release, vendoring

ROOT = Path(".").parent.parent.absolute()

ns = invoke.Collection(vendoring, release, release.clean_mdchangelog)



import json
from collections import defaultdict

import pytest

from pipenv.project import Project
from pipenv.utils import requirements


@pytest.fixture
def pypi_lockfile():
    lockfile = defaultdict(dict)
    lockfile["_meta"] = {
        "sources": [
            {
                "name": "pypi",
                "url": "https://pypi.org/simple",
                "verify_ssl": True,
            }
        ]
    }
    yield lockfile


def test_git_branch_contains_slashes(pipenv_instance_pypi, pypi_lockfile):
    pypi_lockfile["default"]["google-api-python-client"] = {
        "git": "https://github.com/thehesiod/google-api-python-client.git@thehesiod/batch-retries2",
        "markers": "python_version >= '3.7'",
        "ref": "03803c21fc13a345e978f32775b2f2fa23c8e706",
    }

    with pipenv_instance_pypi() as p:
        with open(p.lockfile_path, "w") as f:
            json.dump(pypi_lockfile, f)

        project = Project()
        lockfile = project.load_lockfile(expand_env_vars=False)
        deps = lockfile["default"]
        pip_installable_lines = requirements.requirements_from_lockfile(
            deps, include_hashes=False, include_markers=True
        )
        assert pip_installable_lines == [
            "google-api-python-client @ git+https://github.com/thehesiod/google-api-python-client.git@03803c21fc13a345e978f32775b2f2fa23c8e706"
        ]


def test_git_branch_contains_subdirectory_fragment(pipenv_instance_pypi, pypi_lockfile):
    pypi_lockfile["default"]["pep508_package"] = {
        "git": "https://github.com/techalchemy/test-project.git@master",
        "subdirectory": "parent_folder/pep508-package",
        "ref": "03803c21fc13a345e978f32775b2f2fa23c8e706",
    }

    with pipenv_instance_pypi() as p:
        with open(p.lockfile_path, "w") as f:
            json.dump(pypi_lockfile, f)

        project = Project()
        lockfile = project.load_lockfile(expand_env_vars=False)
        deps = lockfile["default"]
        pip_installable_lines = requirements.requirements_from_lockfile(
            deps, include_hashes=False, include_markers=True
        )
        assert pip_installable_lines == [
            "pep508_package @ git+https://github.com/techalchemy/test-project.git@03803c21fc13a345e978f32775b2f2fa23c8e706#subdirectory=parent_folder/pep508-package"
        ]



import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from pipenv.utils.constants import FALSE_VALUES, TRUE_VALUES
from pipenv.utils.shell import temp_environ


@pytest.mark.dotvenv
@pytest.mark.parametrize("true_value", TRUE_VALUES)
def test_venv_in_project(true_value, pipenv_instance_pypi):
    with temp_environ():
        os.environ["PIPENV_VENV_IN_PROJECT"] = true_value
        with pipenv_instance_pypi() as p:
            c = p.pipenv("install dataclasses-json")
            assert c.returncode == 0
            assert p.path in p.pipenv("--venv").stdout


@pytest.mark.dotvenv
@pytest.mark.parametrize("false_value", FALSE_VALUES)
def test_venv_in_project_disabled_ignores_venv(false_value, pipenv_instance_pypi):
    venv_name = "my_project"
    with temp_environ():
        os.environ["PIPENV_VENV_IN_PROJECT"] = false_value
        with pipenv_instance_pypi() as p:
            file_path = os.path.join(p.path, ".venv")
            with open(file_path, "w") as f:
                f.write(venv_name)

            with temp_environ(), TemporaryDirectory(
                prefix="pipenv-", suffix="temp_workon_home"
            ) as workon_home:
                os.environ["WORKON_HOME"] = workon_home
                c = p.pipenv("install dataclasses-json")
                assert c.returncode == 0
                c = p.pipenv("--venv")
                assert c.returncode == 0
                venv_loc = Path(c.stdout.strip()).resolve()
                assert venv_loc.exists()
                assert venv_loc.joinpath(".project").exists()
                venv_path = Path(venv_loc).resolve()
                venv_expected_path = Path(workon_home).joinpath(venv_name).resolve()
                assert os.path.samefile(venv_path, venv_expected_path)


@pytest.mark.dotvenv
@pytest.mark.parametrize("true_value", TRUE_VALUES)
def test_venv_at_project_root(true_value, pipenv_instance_pypi):
    with temp_environ(), pipenv_instance_pypi() as p:
        os.environ["PIPENV_VENV_IN_PROJECT"] = true_value
        c = p.pipenv("install")
        assert c.returncode == 0
        assert p.path in p.pipenv("--venv").stdout
        del os.environ["PIPENV_VENV_IN_PROJECT"]
        os.mkdir("subdir")
        os.chdir("subdir")
        # should still detect installed
        assert p.path in p.pipenv("--venv").stdout


@pytest.mark.dotvenv
@pytest.mark.parametrize("false_value", FALSE_VALUES)
def test_venv_in_project_disabled_with_existing_venv_dir(
    false_value, pipenv_instance_pypi
):
    venv_name = "my_project"
    with temp_environ(), pipenv_instance_pypi() as p, TemporaryDirectory(
        prefix="pipenv-", suffix="temp_workon_home"
    ) as workon_home:
        os.environ["PIPENV_VENV_IN_PROJECT"] = false_value
        os.environ["PIPENV_CUSTOM_VENV_NAME"] = venv_name
        os.environ["WORKON_HOME"] = workon_home
        os.mkdir(".venv")
        c = p.pipenv("install")
        assert c.returncode == 0
        c = p.pipenv("--venv")
        assert c.returncode == 0
        venv_loc = Path(c.stdout.strip()).resolve()
        assert venv_loc.exists()
        assert venv_loc.joinpath(".project").exists()
        venv_path = Path(venv_loc).resolve()
        venv_expected_path = Path(workon_home).joinpath(venv_name).resolve()
        assert os.path.samefile(venv_path, venv_expected_path)


@pytest.mark.dotvenv
def test_reuse_previous_venv(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        os.mkdir(".venv")
        c = p.pipenv("install dataclasses-json")
        assert c.returncode == 0
        assert p.path in p.pipenv("--venv").stdout


@pytest.mark.dotvenv
@pytest.mark.parametrize("venv_name", ("test-venv", os.path.join("foo", "test-venv")))
def test_venv_file(venv_name, pipenv_instance_pypi):
    """Tests virtualenv creation when a .venv file exists at the project root
    and contains a venv name.
    """
    with pipenv_instance_pypi() as p:
        file_path = os.path.join(p.path, ".venv")
        with open(file_path, "w") as f:
            f.write(venv_name)

        with temp_environ(), TemporaryDirectory(
            prefix="pipenv-", suffix="temp_workon_home"
        ) as workon_home:
            os.environ["WORKON_HOME"] = workon_home

            c = p.pipenv("install")
            assert c.returncode == 0

            c = p.pipenv("--venv")
            assert c.returncode == 0
            venv_loc = Path(c.stdout.strip()).resolve()
            assert venv_loc.exists()
            assert venv_loc.joinpath(".project").exists()
            venv_path = Path(venv_loc).resolve()
            if os.path.sep in venv_name:
                venv_expected_path = Path(p.path).joinpath(venv_name)
            else:
                venv_expected_path = Path(workon_home).joinpath(venv_name)
            assert venv_path == venv_expected_path.resolve()


@pytest.mark.dotvenv
def test_empty_venv_file(pipenv_instance_pypi):
    """Tests virtualenv creation when an empty .venv file exists at the project root"""
    with pipenv_instance_pypi() as p:
        file_path = os.path.join(p.path, ".venv")
        with open(file_path, "w"):
            pass

        with temp_environ(), TemporaryDirectory(
            prefix="pipenv-", suffix="temp_workon_home"
        ) as workon_home:
            os.environ["WORKON_HOME"] = workon_home

            c = p.pipenv("install")
            assert c.returncode == 0

            c = p.pipenv("--venv")
            assert c.returncode == 0
            venv_loc = Path(c.stdout.strip()).absolute()
            assert venv_loc.exists()
            assert venv_loc.joinpath(".project").exists()
            venv_path = Path(venv_loc)
            venv_path_parent = Path(venv_path.parent)
            assert venv_path_parent == Path(workon_home)


@pytest.mark.dotvenv
def test_venv_in_project_default_when_venv_exists(pipenv_instance_pypi):
    """Tests virtualenv creation when a .venv file exists at the project root."""
    with temp_environ(), pipenv_instance_pypi() as p, TemporaryDirectory(
        prefix="pipenv-", suffix="-test_venv"
    ) as venv_path:
        file_path = os.path.join(p.path, ".venv")
        with open(file_path, "w") as f:
            f.write(venv_path)

        c = p.pipenv("install")
        assert c.returncode == 0
        c = p.pipenv("--venv")
        assert c.returncode == 0
        venv_loc = Path(c.stdout.strip())

        assert venv_loc.joinpath(".project").exists()
        assert venv_loc == Path(venv_path)


@pytest.mark.dotenv
def test_venv_name_accepts_custom_name_environment_variable(pipenv_instance_pypi):
    """Tests that virtualenv reads PIPENV_CUSTOM_VENV_NAME and accepts it as a name"""
    with pipenv_instance_pypi() as p:
        test_name = "sensible_custom_venv_name"
        with temp_environ():
            os.environ["PIPENV_CUSTOM_VENV_NAME"] = test_name
            c = p.pipenv("install")
            assert c.returncode == 0
            c = p.pipenv("--venv")
            assert c.returncode == 0
            venv_path = c.stdout.strip()
            assert test_name == Path(venv_path).parts[-1]



import ast
import os

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

ROOT = os.path.dirname(__file__)

PACKAGE_NAME = "legacy_backend_package"

VERSION = None

with open(os.path.join(ROOT, "src", PACKAGE_NAME.replace("-", "_"), "__init__.py")) as f:
    for line in f:
        if line.startswith("__version__ = "):
            VERSION = ast.literal_eval(line[len("__version__ = ") :].strip())
            break
if VERSION is None:
    raise OSError("failed to read version")


# Put everything in setup.cfg, except those that don't actually work?
setup(
    # These really don't work.
    package_dir={"": "src"},
    packages=find_packages("src"),
    # I don't know how to specify an empty key in setup.cfg.
    package_data={
        "": ["LICENSE*", "README*"],
    },
    # I need this to be dynamic.
    version=VERSION,
)



"""
This code wraps the vendored appdirs module to so the return values are
compatible for the current pip code base.

The intention is to rewrite current usages gradually, keeping the tests pass,
and eventually drop this after all usages are changed.
"""
from __future__ import annotations

import os
import sys

from pipenv.patched.pip._vendor import platformdirs as _appdirs


def user_cache_dir(appname: str) -> str:
    return _appdirs.user_cache_dir(appname, appauthor=False)


def _macos_user_config_dir(appname: str, roaming: bool = True) -> str:
    # Use ~/Application Support/pip, if the directory exists.
    path = _appdirs.user_data_dir(appname, appauthor=False, roaming=roaming)
    if os.path.isdir(path):
        return path

    # Use a Linux-like ~/.config/pip, by default.
    linux_like_path = "~/.config/"
    if appname:
        linux_like_path = os.path.join(linux_like_path, appname)

    return os.path.expanduser(linux_like_path)


def user_config_dir(appname: str, roaming: bool = True) -> str:
    if sys.platform == "darwin":
        return _macos_user_config_dir(appname, roaming)

    return _appdirs.user_config_dir(appname, appauthor=False, roaming=roaming)


# for the discussion regarding site_config_dir locations
# see <https://github.com/pypa/pip/issues/1733>
def site_config_dirs(appname: str) -> list[str]:
    if sys.platform == "darwin":
        dirval = _appdirs.site_data_dir(appname, appauthor=False, multipath=True)
        return dirval.split(os.pathsep)

    dirval = _appdirs.site_config_dir(appname, appauthor=False, multipath=True)
    if sys.platform == "win32":
        return [dirval]

    # Unix-y system. Look in /etc as well.
    return dirval.split(os.pathsep) + ["/etc"]



import os
import shutil
from pathlib import Path

import pytest

from pipenv.project import Project
from pipenv.utils.pylock import PylockFile, find_pylock_file


@pytest.fixture
def pylock_project(tmp_path):
    """Create a temporary project with a pylock.toml file."""
    # Copy the example pylock.toml to the temporary directory
    example_pylock = Path(__file__).parent.parent.parent / "examples" / "pylock.toml"
    tmp_pylock = tmp_path / "pylock.toml"

    # Create a simple Pipfile
    pipfile_content = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
requests = "*"

[dev-packages]

[requires]
python_version = "3.8"
"""

    with open(tmp_path / "Pipfile", "w") as f:
        f.write(pipfile_content)

    shutil.copy(example_pylock, tmp_pylock)

    # Change to the temporary directory
    old_cwd = os.getcwd()
    os.chdir(tmp_path)

    try:
        yield tmp_path
    finally:
        os.chdir(old_cwd)


def test_find_pylock_file(pylock_project):
    """Test that find_pylock_file correctly finds the pylock.toml file."""
    pylock_path = find_pylock_file(pylock_project)
    assert pylock_path is not None
    assert pylock_path.name == "pylock.toml"
    assert pylock_path.exists()


def test_pylock_file_loading(pylock_project):
    """Test loading a pylock.toml file."""
    pylock_path = pylock_project / "pylock.toml"
    pylock = PylockFile.from_path(pylock_path)

    assert pylock.lock_version == "1.0"
    assert pylock.created_by == "pipenv"
    assert pylock.requires_python == ">=3.8"
    # Updated example now has 5 packages: requests, urllib3, certifi, pytest, cryptography
    assert len(pylock.packages) == 5
    assert pylock.packages[0]["name"] == "requests"
    assert pylock.packages[0]["version"] == "2.28.1"
    # Check new index field
    assert pylock.packages[0].get("index") == "https://pypi.org/simple/"


def test_project_pylock_integration(pylock_project):
    """Test that Project class correctly detects and uses pylock.toml."""
    # Create a project instance
    project = Project(chdir=False)

    # Check that pylock.toml is detected
    assert project.pylock_exists
    assert project.pylock_location is not None
    assert Path(project.pylock_location).name == "pylock.toml"

    # Check that lockfile_content returns the converted pylock content
    lockfile_content = project.lockfile_content
    assert "_meta" in lockfile_content
    assert "default" in lockfile_content
    assert "requests" in lockfile_content["default"]
    assert "urllib3" in lockfile_content["default"]
    assert "certifi" in lockfile_content["default"]
    # pytest should be in develop section due to dependency_groups marker
    assert "develop" in lockfile_content
    assert "pytest" in lockfile_content["develop"]

    # Check that the converted content has the correct format
    requests_entry = lockfile_content["default"]["requests"]
    assert requests_entry["version"] == "==2.28.1"
    assert "hashes" in requests_entry
    assert len(requests_entry["hashes"]) == 1
    assert requests_entry["hashes"][0].startswith("sha256:")


@pytest.fixture
def pylock_write_project(tmp_path):
    """Create a temporary project with a Pipfile that has use_pylock enabled."""
    # Create a simple Pipfile with use_pylock enabled
    pipfile_content = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
requests = "*"

[dev-packages]

[requires]
python_version = "3.8"

[pipenv]
use_pylock = true
"""

    # Create a simple Pipfile.lock
    lockfile_content = """
{
    "_meta": {
        "hash": {
            "sha256": "b8c2e1580c53e383cfe4254c1f16560b855d984c674dc07bcce19a8b5b28c6b2"
        },
        "pipfile-spec": 6,
        "requires": {
            "python_version": "3.8"
        },
        "sources": [
            {
                "name": "pypi",
                "url": "https://pypi.org/simple",
                "verify_ssl": true
            }
        ]
    },
    "default": {
        "certifi": {
            "hashes": [
                "sha256:0d9c601124e5a6ba9712dbc60d9c53c21e34f5f641fe83002317394311bdce14"
            ],
            "version": "==2022.9.24"
        },
        "charset-normalizer": {
            "hashes": [
                "sha256:83e9a75d1911279afd89352c68b45348559d1fc0506b054b346651b5e7fee29f"
            ],
            "version": "==2.1.1"
        },
        "idna": {
            "hashes": [
                "sha256:90b77e79eaa3eba6de819a0c442c0b4ceefc341a7a2ab77d7562bf49f425c5c2"
            ],
            "version": "==3.4"
        },
        "requests": {
            "hashes": [
                "sha256:b8aa58f8cf793ffd8782d3d8cb19e66ef36f7aba4353eec859e74678b01b07a7"
            ],
            "index": "pypi",
            "version": "==2.28.1"
        },
        "urllib3": {
            "hashes": [
                "sha256:b930dd878d5a8afb066a637fbb35144fe7901e3b209d1cd4f524bd0e9deee997"
            ],
            "version": "==1.26.12"
        }
    },
    "develop": {}
}
"""

    with open(tmp_path / "Pipfile", "w") as f:
        f.write(pipfile_content)

    with open(tmp_path / "Pipfile.lock", "w") as f:
        f.write(lockfile_content)

    # Change to the temporary directory
    old_cwd = os.getcwd()
    os.chdir(tmp_path)

    try:
        yield tmp_path
    finally:
        os.chdir(old_cwd)


def test_write_pylock_file(pylock_write_project):
    """Test that Project class correctly writes pylock.toml files."""
    # Create a project instance
    project = Project(chdir=False)

    # Check that use_pylock is enabled
    assert project.use_pylock is True

    # Check that pylock_output_path is correct
    assert project.pylock_output_path == str(pylock_write_project / "pylock.toml")

    # Load the lockfile content
    lockfile_content = project.lockfile_content

    # Write the lockfile (which should also write pylock.toml)
    project.write_lockfile(lockfile_content)

    # Check that pylock.toml was created
    pylock_path = pylock_write_project / "pylock.toml"
    assert pylock_path.exists()

    # Load the pylock.toml file and verify its contents
    pylock = PylockFile.from_path(pylock_path)

    # Check basic properties
    assert pylock.lock_version == "1.0"
    assert pylock.created_by == "pipenv"

    # Check that all packages are included
    package_names = [p["name"] for p in pylock.packages]
    assert "requests" in package_names
    assert "urllib3" in package_names
    assert "certifi" in package_names
    assert "charset-normalizer" in package_names
    assert "idna" in package_names

    # Check that the tool.pipenv section exists
    assert "pipenv" in pylock.tool
    assert "generated_from" in pylock.tool["pipenv"]
    assert pylock.tool["pipenv"]["generated_from"] == "Pipfile.lock"


@pytest.fixture
def pylock_write_named_project(tmp_path):
    """Create a temporary project with a Pipfile that has use_pylock and pylock_name enabled."""
    # Create a simple Pipfile with use_pylock and pylock_name enabled
    pipfile_content = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
requests = "*"

[dev-packages]

[requires]
python_version = "3.8"

[pipenv]
use_pylock = true
pylock_name = "dev"
"""

    # Create a simple Pipfile.lock (same as in pylock_write_project)
    lockfile_content = """
{
    "_meta": {
        "hash": {
            "sha256": "b8c2e1580c53e383cfe4254c1f16560b855d984c674dc07bcce19a8b5b28c6b2"
        },
        "pipfile-spec": 6,
        "requires": {
            "python_version": "3.8"
        },
        "sources": [
            {
                "name": "pypi",
                "url": "https://pypi.org/simple",
                "verify_ssl": true
            }
        ]
    },
    "default": {
        "certifi": {
            "hashes": [
                "sha256:0d9c601124e5a6ba9712dbc60d9c53c21e34f5f641fe83002317394311bdce14"
            ],
            "version": "==2022.9.24"
        },
        "charset-normalizer": {
            "hashes": [
                "sha256:83e9a75d1911279afd89352c68b45348559d1fc0506b054b346651b5e7fee29f"
            ],
            "version": "==2.1.1"
        },
        "idna": {
            "hashes": [
                "sha256:90b77e79eaa3eba6de819a0c442c0b4ceefc341a7a2ab77d7562bf49f425c5c2"
            ],
            "version": "==3.4"
        },
        "requests": {
            "hashes": [
                "sha256:b8aa58f8cf793ffd8782d3d8cb19e66ef36f7aba4353eec859e74678b01b07a7"
            ],
            "index": "pypi",
            "version": "==2.28.1"
        },
        "urllib3": {
            "hashes": [
                "sha256:b930dd878d5a8afb066a637fbb35144fe7901e3b209d1cd4f524bd0e9deee997"
            ],
            "version": "==1.26.12"
        }
    },
    "develop": {}
}
"""

    with open(tmp_path / "Pipfile", "w") as f:
        f.write(pipfile_content)

    with open(tmp_path / "Pipfile.lock", "w") as f:
        f.write(lockfile_content)

    # Change to the temporary directory
    old_cwd = os.getcwd()
    os.chdir(tmp_path)

    try:
        yield tmp_path
    finally:
        os.chdir(old_cwd)


def test_write_named_pylock_file(pylock_write_named_project):
    """Test that Project class correctly writes named pylock.toml files."""
    # Create a project instance
    project = Project(chdir=False)

    # Check that use_pylock is enabled
    assert project.use_pylock is True

    # Check that pylock_name is set
    assert project.settings.get("pylock_name") == "dev"

    # Check that pylock_output_path is correct
    assert project.pylock_output_path == str(pylock_write_named_project / "pylock.dev.toml")

    # Load the lockfile content
    lockfile_content = project.lockfile_content

    # Write the lockfile (which should also write pylock.dev.toml)
    project.write_lockfile(lockfile_content)

    # Check that pylock.dev.toml was created
    pylock_path = pylock_write_named_project / "pylock.dev.toml"
    assert pylock_path.exists()

    # Load the pylock.dev.toml file and verify its contents
    pylock = PylockFile.from_path(pylock_path)

    # Check basic properties
    assert pylock.lock_version == "1.0"
    assert pylock.created_by == "pipenv"

    # Check that all packages are included
    package_names = [p["name"] for p in pylock.packages]
    assert "requests" in package_names
    assert "urllib3" in package_names
    assert "certifi" in package_names
    assert "charset-normalizer" in package_names
    assert "idna" in package_names



import sys

import pytest

from pipenv.utils.shell import temp_environ

from .conftest import DEFAULT_PRIVATE_PYPI_SERVER


@pytest.mark.uninstall
@pytest.mark.install
def test_uninstall_requests(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        c = p.pipenv("install requests")
        assert c.returncode == 0
        assert "requests" in p.pipfile["packages"]

        c = p.pipenv("uninstall requests")
        assert c.returncode == 0
        assert "requests" not in p.pipfile["packages"]
        assert "requests" not in p.lockfile["default"]


@pytest.mark.uninstall
@pytest.mark.skipif(
    sys.version_info >= (3, 12), reason="Package does not work with Python 3.12"
)
def test_uninstall_django(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        c = p.pipenv("install Django")
        assert c.returncode == 0
        assert "django" in p.pipfile["packages"]
        assert "django" in p.lockfile["default"]
        assert "pytz" in p.lockfile["default"]

        c = p.pipenv("run python -m django --version")
        assert c.returncode == 0

        c = p.pipenv("uninstall Django")
        assert c.returncode == 0
        assert "django" not in p.pipfile["dev-packages"]
        assert "django" not in p.lockfile["develop"]
        assert p.lockfile["develop"] == {}

        c = p.pipenv("run python -m django --version")
        assert c.returncode > 0


@pytest.mark.install
@pytest.mark.uninstall
@pytest.mark.skipif(
    sys.version_info >= (3, 12), reason="Package does not work with Python 3.12"
)
def test_mirror_uninstall(pipenv_instance_pypi):
    with temp_environ(), pipenv_instance_pypi() as p:
        mirror_url = DEFAULT_PRIVATE_PYPI_SERVER
        assert "pypi.org" not in mirror_url

        c = p.pipenv(f"install Django --pypi-mirror {mirror_url}")
        assert c.returncode == 0
        assert "django" in p.pipfile["packages"]
        assert "django" in p.lockfile["default"]
        assert "pytz" in p.lockfile["default"]
        # Ensure the --pypi-mirror parameter hasn't altered the Pipfile or Pipfile.lock sources
        assert len(p.pipfile["source"]) == 1
        assert len(p.lockfile["_meta"]["sources"]) == 1
        assert p.pipfile["source"][0]["url"] == "https://pypi.org/simple"
        assert p.lockfile["_meta"]["sources"][0]["url"] == "https://pypi.org/simple"

        c = p.pipenv("run python -m django --version")
        assert c.returncode == 0

        c = p.pipenv(f"uninstall Django --pypi-mirror {mirror_url}")
        assert c.returncode == 0
        assert "django" not in p.pipfile["dev-packages"]
        assert "django" not in p.lockfile["develop"]
        assert p.lockfile["develop"] == {}
        # Ensure the --pypi-mirror parameter hasn't altered the Pipfile or Pipfile.lock sources
        assert len(p.pipfile["source"]) == 1
        assert len(p.lockfile["_meta"]["sources"]) == 1
        assert p.pipfile["source"][0]["url"] == "https://pypi.org/simple"
        assert p.lockfile["_meta"]["sources"][0]["url"] == "https://pypi.org/simple"

        c = p.pipenv("run python -m django --version")
        assert c.returncode > 0


@pytest.mark.files
@pytest.mark.install
@pytest.mark.uninstall
def test_uninstall_all_local_files(pipenv_instance_private_pypi, testsroot):
    with pipenv_instance_private_pypi() as p:
        file_uri = p._pipfile.get_fixture_path(
            "tablib/tablib-0.12.1.tar.gz", fixtures="pypi"
        ).as_uri()
        c = p.pipenv(f"install {file_uri}")
        assert c.returncode == 0
        c = p.pipenv("uninstall --all")
        assert c.returncode == 0
        assert "tablib" in c.stdout
        # Uninstall --all is not supposed to remove things from the pipfile
        # Note that it didn't before, but that instead local filenames showed as hashes
        assert "tablib" in p.pipfile["packages"]


@pytest.mark.install
@pytest.mark.uninstall
def test_uninstall_all_dev(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = f"""
[[source]]
name = "pypi"
url = "{p.index_url}"
verify_ssl = true

[packages]
tablib = "*"

[dev-packages]
jinja2 = "==2.11.1"
six = "==1.12.0"
        """
            f.write(contents)

        c = p.pipenv("install -v --dev")
        assert c.returncode == 0

        assert "tablib" in p.pipfile["packages"]
        assert "jinja2" in p.pipfile["dev-packages"]
        assert "six" in p.pipfile["dev-packages"]
        assert "tablib" in p.lockfile["default"]
        assert "jinja2" in p.lockfile["develop"]
        assert "six" in p.lockfile["develop"]
        assert c.returncode == 0

        c = p.pipenv("uninstall -v --all-dev")
        assert c.returncode == 0
        assert p.pipfile["dev-packages"] == {}
        assert "jinja2" not in p.lockfile["develop"]
        assert "six" not in p.lockfile["develop"]
        assert "tablib" in p.pipfile["packages"]
        assert "tablib" in p.lockfile["default"]

        c = p.pipenv('run python -c "import jinja2"')
        assert c.returncode > 0

        c = p.pipenv('run python -c "import tablib"')
        assert c.returncode == 0


@pytest.mark.install
@pytest.mark.uninstall
def test_normalize_name_uninstall(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
# Pre comment
[packages]
Requests = "*"
python_DateUtil = "*"
"""
            f.write(contents)

        c = p.pipenv("install")
        assert c.returncode == 0

        c = p.pipenv("uninstall python_dateutil")
        assert "Requests" in p.pipfile["packages"]
        assert "python_DateUtil" not in p.pipfile["packages"]
        with open(p.pipfile_path) as f:
            contents = f.read()
            assert "# Pre comment" in contents


@pytest.mark.install
@pytest.mark.uninstall
def test_uninstall_all_dev_with_shared_dependencies(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
        [packages]
        pytest = "==4.6.11"

        [dev-packages]
        six = "*"
        """
            f.write(contents)
        c = p.pipenv("install")
        assert c.returncode == 0

        c = p.pipenv("uninstall --all-dev")
        assert c.returncode == 0

        assert "six" in p.lockfile["default"]


@pytest.mark.install
@pytest.mark.uninstall
def test_uninstall_missing_parameters(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        c = p.pipenv("install six")
        assert c.returncode == 0

        c = p.pipenv("uninstall")
        assert c.returncode != 0
        assert "No package provided!" in c.stderr


@pytest.mark.categories
@pytest.mark.install
@pytest.mark.uninstall
def test_uninstall_category_with_shared_requirement(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
        [packages]
        six = "*"

        [prereq]
        six = "*"
        """
            f.write(contents)
        c = p.pipenv("install")
        assert c.returncode == 0

        c = p.pipenv("uninstall six --categories default")
        assert c.returncode == 0

        assert "six" in p.lockfile["prereq"]
        assert "six" not in p.lockfile["default"]


@pytest.mark.categories
@pytest.mark.install
@pytest.mark.uninstall
def test_uninstall_multiple_categories(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
        [after]
        six = "==1.12.0"

        [prereq]
        six = "==1.12.0"
        """
            f.write(contents)
        c = p.pipenv("install")
        assert c.returncode == 0

        c = p.pipenv('uninstall six --categories="prereq after"')
        assert c.returncode == 0

        assert "six" not in p.lockfile.get("prereq", {})
        assert "six" not in p.lockfile["default"]


@pytest.mark.install
@pytest.mark.uninstall
def test_category_sorted_alphabetically_with_directive(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[pipenv]
sort_pipfile = true

[packages]
parse = "*"
colorama = "*"
build = "*"
atomicwrites = "*"
            """.strip()
            f.write(contents)

        c = p.pipenv("install")
        assert c.returncode == 0

        c = p.pipenv("uninstall build")
        assert c.returncode == 0
        assert "build" not in p.pipfile["packages"]
        assert list(p.pipfile["packages"].keys()) == ["atomicwrites", "colorama", "parse"]


@pytest.mark.install
@pytest.mark.uninstall
def test_sorting_handles_str_values_and_dict_values(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[pipenv]
sort_pipfile = true

[packages]
zipp = "*"
parse = {version = "*"}
colorama = "*"
build = "*"
atomicwrites = {version = "*"}
            """.strip()
            f.write(contents)
        c = p.pipenv("install")
        assert c.returncode == 0

        c = p.pipenv("uninstall build")
        assert c.returncode == 0
        assert "build" not in p.pipfile["packages"]
        assert list(p.pipfile["packages"].keys()) == [
            "atomicwrites",
            "colorama",
            "parse",
            "zipp",
        ]


@pytest.mark.install
@pytest.mark.uninstall
def test_category_not_sorted_without_directive(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
parse = "*"
colorama = "*"
build = "*"
atomicwrites = "*"
            """.strip()
            f.write(contents)

        c = p.pipenv("install")
        assert c.returncode == 0

        c = p.pipenv("uninstall build")
        assert c.returncode == 0
        assert "build" not in p.pipfile["packages"]
        assert list(p.pipfile["packages"].keys()) == [
            "parse",
            "colorama",
            "atomicwrites",
        ]


@pytest.mark.uninstall
def test_uninstall_without_venv(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
colorama = "*"
atomicwrites = "*"
            """.strip()
            f.write(contents)

        c = p.pipenv("install")
        assert c.returncode == 0

        c = p.pipenv("uninstall --all")
        assert c.returncode == 0
        # uninstall --all shold not remove packages from Pipfile
        assert list(p.pipfile["packages"].keys()) == ["colorama", "atomicwrites"]



"""Stuff that differs in different Python versions and platform
distributions."""

import importlib.resources
import logging
import os
import sys
from typing import IO

__all__ = ["get_path_uid", "stdlib_pkgs", "tomllib", "WINDOWS"]


logger = logging.getLogger(__name__)


def has_tls() -> bool:
    try:
        import _ssl  # noqa: F401  # ignore unused

        return True
    except ImportError:
        pass

    from pipenv.patched.pip._vendor.urllib3.util import IS_PYOPENSSL

    return IS_PYOPENSSL


def get_path_uid(path: str) -> int:
    """
    Return path's uid.

    Does not follow symlinks:
        https://github.com/pypa/pip/pull/935#discussion_r5307003

    Placed this function in compat due to differences on AIX and
    Jython, that should eventually go away.

    :raises OSError: When path is a symlink or can't be read.
    """
    if hasattr(os, "O_NOFOLLOW"):
        fd = os.open(path, os.O_RDONLY | os.O_NOFOLLOW)
        file_uid = os.fstat(fd).st_uid
        os.close(fd)
    else:  # AIX and Jython
        # WARNING: time of check vulnerability, but best we can do w/o NOFOLLOW
        if not os.path.islink(path):
            # older versions of Jython don't have `os.fstat`
            file_uid = os.stat(path).st_uid
        else:
            # raise OSError for parity with os.O_NOFOLLOW above
            raise OSError(f"{path} is a symlink; Will not return uid for symlinks")
    return file_uid


# The importlib.resources.open_text function was deprecated in 3.11 with suggested
# replacement we use below.
if sys.version_info < (3, 11):
    open_text_resource = importlib.resources.open_text
else:

    def open_text_resource(
        package: str, resource: str, encoding: str = "utf-8", errors: str = "strict"
    ) -> IO[str]:
        return (importlib.resources.files(package) / resource).open(
            "r", encoding=encoding, errors=errors
        )


if sys.version_info >= (3, 11):
    import tomllib
else:
    from pipenv.patched.pip._vendor import tomli as tomllib


# packages in the stdlib that may have installation metadata, but should not be
# considered 'installed'.  this theoretically could be determined based on
# dist.location (py27:`sysconfig.get_paths()['stdlib']`,
# py26:sysconfig.get_config_vars('LIBDEST')), but fear platform variation may
# make this ineffective, so hard-coding
stdlib_pkgs = {"python", "wsgiref", "argparse"}


# windows detection, covers cpython and ironpython
WINDOWS = sys.platform.startswith("win") or (sys.platform == "cli" and os.name == "nt")



import io
import logging
import os
import pathlib
import shutil
import stat
import sys
import tempfile
from collections import OrderedDict
from contextlib import contextmanager
from typing import IO, Dict, Iterable, Iterator, Mapping, Optional, Tuple, Union

from .parser import Binding, parse_stream
from .variables import parse_variables

# A type alias for a string path to be used for the paths in this file.
# These paths may flow to `open()` and `shutil.move()`; `shutil.move()`
# only accepts string paths, not byte paths or file descriptors. See
# https://github.com/python/typeshed/pull/6832.
StrPath = Union[str, "os.PathLike[str]"]

logger = logging.getLogger(__name__)


def _load_dotenv_disabled() -> bool:
    """
    Determine if dotenv loading has been disabled.
    """
    if "PYTHON_DOTENV_DISABLED" not in os.environ:
        return False
    value = os.environ["PYTHON_DOTENV_DISABLED"].casefold()
    return value in {"1", "true", "t", "yes", "y"}


def with_warn_for_invalid_lines(mappings: Iterator[Binding]) -> Iterator[Binding]:
    for mapping in mappings:
        if mapping.error:
            logger.warning(
                "python-dotenv could not parse statement starting at line %s",
                mapping.original.line,
            )
        yield mapping


class DotEnv:
    def __init__(
        self,
        dotenv_path: Optional[StrPath],
        stream: Optional[IO[str]] = None,
        verbose: bool = False,
        encoding: Optional[str] = None,
        interpolate: bool = True,
        override: bool = True,
    ) -> None:
        self.dotenv_path: Optional[StrPath] = dotenv_path
        self.stream: Optional[IO[str]] = stream
        self._dict: Optional[Dict[str, Optional[str]]] = None
        self.verbose: bool = verbose
        self.encoding: Optional[str] = encoding
        self.interpolate: bool = interpolate
        self.override: bool = override

    @contextmanager
    def _get_stream(self) -> Iterator[IO[str]]:
        if self.dotenv_path and _is_file_or_fifo(self.dotenv_path):
            with open(self.dotenv_path, encoding=self.encoding) as stream:
                yield stream
        elif self.stream is not None:
            yield self.stream
        else:
            if self.verbose:
                logger.info(
                    "python-dotenv could not find configuration file %s.",
                    self.dotenv_path or ".env",
                )
            yield io.StringIO("")

    def dict(self) -> Dict[str, Optional[str]]:
        """Return dotenv as dict"""
        if self._dict:
            return self._dict

        raw_values = self.parse()

        if self.interpolate:
            self._dict = OrderedDict(
                resolve_variables(raw_values, override=self.override)
            )
        else:
            self._dict = OrderedDict(raw_values)

        return self._dict

    def parse(self) -> Iterator[Tuple[str, Optional[str]]]:
        with self._get_stream() as stream:
            for mapping in with_warn_for_invalid_lines(parse_stream(stream)):
                if mapping.key is not None:
                    yield mapping.key, mapping.value

    def set_as_environment_variables(self) -> bool:
        """
        Load the current dotenv as system environment variable.
        """
        if not self.dict():
            return False

        for k, v in self.dict().items():
            if k in os.environ and not self.override:
                continue
            if v is not None:
                os.environ[k] = v

        return True

    def get(self, key: str) -> Optional[str]:
        """ """
        data = self.dict()

        if key in data:
            return data[key]

        if self.verbose:
            logger.warning("Key %s not found in %s.", key, self.dotenv_path)

        return None


def get_key(
    dotenv_path: StrPath,
    key_to_get: str,
    encoding: Optional[str] = "utf-8",
) -> Optional[str]:
    """
    Get the value of a given key from the given .env.

    Returns `None` if the key isn't found or doesn't have a value.
    """
    return DotEnv(dotenv_path, verbose=True, encoding=encoding).get(key_to_get)


@contextmanager
def rewrite(
    path: StrPath,
    encoding: Optional[str],
) -> Iterator[Tuple[IO[str], IO[str]]]:
    pathlib.Path(path).touch()

    with tempfile.NamedTemporaryFile(mode="w", encoding=encoding, delete=False) as dest:
        error = None
        try:
            with open(path, encoding=encoding) as source:
                yield (source, dest)
        except BaseException as err:
            error = err

    if error is None:
        shutil.move(dest.name, path)
    else:
        os.unlink(dest.name)
        raise error from None


def set_key(
    dotenv_path: StrPath,
    key_to_set: str,
    value_to_set: str,
    quote_mode: str = "always",
    export: bool = False,
    encoding: Optional[str] = "utf-8",
) -> Tuple[Optional[bool], str, str]:
    """
    Adds or Updates a key/value to the given .env

    If the .env path given doesn't exist, fails instead of risking creating
    an orphan .env somewhere in the filesystem
    """
    if quote_mode not in ("always", "auto", "never"):
        raise ValueError(f"Unknown quote_mode: {quote_mode}")

    quote = quote_mode == "always" or (
        quote_mode == "auto" and not value_to_set.isalnum()
    )

    if quote:
        value_out = "'{}'".format(value_to_set.replace("'", "\\'"))
    else:
        value_out = value_to_set
    if export:
        line_out = f"export {key_to_set}={value_out}\n"
    else:
        line_out = f"{key_to_set}={value_out}\n"

    with rewrite(dotenv_path, encoding=encoding) as (source, dest):
        replaced = False
        missing_newline = False
        for mapping in with_warn_for_invalid_lines(parse_stream(source)):
            if mapping.key == key_to_set:
                dest.write(line_out)
                replaced = True
            else:
                dest.write(mapping.original.string)
                missing_newline = not mapping.original.string.endswith("\n")
        if not replaced:
            if missing_newline:
                dest.write("\n")
            dest.write(line_out)

    return True, key_to_set, value_to_set


def unset_key(
    dotenv_path: StrPath,
    key_to_unset: str,
    quote_mode: str = "always",
    encoding: Optional[str] = "utf-8",
) -> Tuple[Optional[bool], str]:
    """
    Removes a given key from the given `.env` file.

    If the .env path given doesn't exist, fails.
    If the given key doesn't exist in the .env, fails.
    """
    if not os.path.exists(dotenv_path):
        logger.warning("Can't delete from %s - it doesn't exist.", dotenv_path)
        return None, key_to_unset

    removed = False
    with rewrite(dotenv_path, encoding=encoding) as (source, dest):
        for mapping in with_warn_for_invalid_lines(parse_stream(source)):
            if mapping.key == key_to_unset:
                removed = True
            else:
                dest.write(mapping.original.string)

    if not removed:
        logger.warning(
            "Key %s not removed from %s - key doesn't exist.", key_to_unset, dotenv_path
        )
        return None, key_to_unset

    return removed, key_to_unset


def resolve_variables(
    values: Iterable[Tuple[str, Optional[str]]],
    override: bool,
) -> Mapping[str, Optional[str]]:
    new_values: Dict[str, Optional[str]] = {}

    for name, value in values:
        if value is None:
            result = None
        else:
            atoms = parse_variables(value)
            env: Dict[str, Optional[str]] = {}
            if override:
                env.update(os.environ)  # type: ignore
                env.update(new_values)
            else:
                env.update(new_values)
                env.update(os.environ)  # type: ignore
            result = "".join(atom.resolve(env) for atom in atoms)

        new_values[name] = result

    return new_values


def _walk_to_root(path: str) -> Iterator[str]:
    """
    Yield directories starting from the given directory up to the root
    """
    if not os.path.exists(path):
        raise IOError("Starting path not found")

    if os.path.isfile(path):
        path = os.path.dirname(path)

    last_dir = None
    current_dir = os.path.abspath(path)
    while last_dir != current_dir:
        yield current_dir
        parent_dir = os.path.abspath(os.path.join(current_dir, os.path.pardir))
        last_dir, current_dir = current_dir, parent_dir


def find_dotenv(
    filename: str = ".env",
    raise_error_if_not_found: bool = False,
    usecwd: bool = False,
) -> str:
    """
    Search in increasingly higher folders for the given file

    Returns path to the file if found, or an empty string otherwise
    """

    def _is_interactive():
        """Decide whether this is running in a REPL or IPython notebook"""
        if hasattr(sys, "ps1") or hasattr(sys, "ps2"):
            return True
        try:
            main = __import__("__main__", None, None, fromlist=["__file__"])
        except ModuleNotFoundError:
            return False
        return not hasattr(main, "__file__")

    def _is_debugger():
        return sys.gettrace() is not None

    if usecwd or _is_interactive() or _is_debugger() or getattr(sys, "frozen", False):
        # Should work without __file__, e.g. in REPL or IPython notebook.
        path = os.getcwd()
    else:
        # will work for .py files
        frame = sys._getframe()
        current_file = __file__

        while frame.f_code.co_filename == current_file or not os.path.exists(
            frame.f_code.co_filename
        ):
            assert frame.f_back is not None
            frame = frame.f_back
        frame_filename = frame.f_code.co_filename
        path = os.path.dirname(os.path.abspath(frame_filename))

    for dirname in _walk_to_root(path):
        check_path = os.path.join(dirname, filename)
        if _is_file_or_fifo(check_path):
            return check_path

    if raise_error_if_not_found:
        raise IOError("File not found")

    return ""


def load_dotenv(
    dotenv_path: Optional[StrPath] = None,
    stream: Optional[IO[str]] = None,
    verbose: bool = False,
    override: bool = False,
    interpolate: bool = True,
    encoding: Optional[str] = "utf-8",
) -> bool:
    """Parse a .env file and then load all the variables found as environment variables.

    Parameters:
        dotenv_path: Absolute or relative path to .env file.
        stream: Text stream (such as `io.StringIO`) with .env content, used if
            `dotenv_path` is `None`.
        verbose: Whether to output a warning the .env file is missing.
        override: Whether to override the system environment variables with the variables
            from the `.env` file.
        encoding: Encoding to be used to read the file.
    Returns:
        Bool: True if at least one environment variable is set else False

    If both `dotenv_path` and `stream` are `None`, `find_dotenv()` is used to find the
    .env file with it's default parameters. If you need to change the default parameters
    of `find_dotenv()`, you can explicitly call `find_dotenv()` and pass the result
    to this function as `dotenv_path`.

    If the environment variable `PYTHON_DOTENV_DISABLED` is set to a truthy value,
    .env loading is disabled.
    """
    if _load_dotenv_disabled():
        logger.debug(
            "python-dotenv: .env loading disabled by PYTHON_DOTENV_DISABLED environment variable"
        )
        return False

    if dotenv_path is None and stream is None:
        dotenv_path = find_dotenv()

    dotenv = DotEnv(
        dotenv_path=dotenv_path,
        stream=stream,
        verbose=verbose,
        interpolate=interpolate,
        override=override,
        encoding=encoding,
    )
    return dotenv.set_as_environment_variables()


def dotenv_values(
    dotenv_path: Optional[StrPath] = None,
    stream: Optional[IO[str]] = None,
    verbose: bool = False,
    interpolate: bool = True,
    encoding: Optional[str] = "utf-8",
) -> Dict[str, Optional[str]]:
    """
    Parse a .env file and return its content as a dict.

    The returned dict will have `None` values for keys without values in the .env file.
    For example, `foo=bar` results in `{"foo": "bar"}` whereas `foo` alone results in
    `{"foo": None}`

    Parameters:
        dotenv_path: Absolute or relative path to the .env file.
        stream: `StringIO` object with .env content, used if `dotenv_path` is `None`.
        verbose: Whether to output a warning if the .env file is missing.
        encoding: Encoding to be used to read the file.

    If both `dotenv_path` and `stream` are `None`, `find_dotenv()` is used to find the
    .env file.
    """
    if dotenv_path is None and stream is None:
        dotenv_path = find_dotenv()

    return DotEnv(
        dotenv_path=dotenv_path,
        stream=stream,
        verbose=verbose,
        interpolate=interpolate,
        override=True,
        encoding=encoding,
    ).dict()


def _is_file_or_fifo(path: StrPath) -> bool:
    """
    Return True if `path` exists and is either a regular file or a FIFO.
    """
    if os.path.isfile(path):
        return True

    try:
        st = os.stat(path)
    except (FileNotFoundError, OSError):
        return False

    return stat.S_ISFIFO(st.st_mode)



import collections
import functools


# from jaraco.functools 4.0.2
def save_method_args(method):
    """
    Wrap a method such that when it is called, the args and kwargs are
    saved on the method.
    """
    args_and_kwargs = collections.namedtuple('args_and_kwargs', 'args kwargs')  # noqa: PYI024

    @functools.wraps(method)
    def wrapper(self, /, *args, **kwargs):
        attr_name = '_saved_' + method.__name__
        attr = args_and_kwargs(args, kwargs)
        setattr(self, attr_name, attr)
        return method(self, *args, **kwargs)

    return wrapper



import pytest


@pytest.mark.install
@pytest.mark.uninstall
def test_uninstall_dev_flag(pipenv_instance_private_pypi):
    """Ensure that running `pipenv uninstall --dev` properly removes packages from dev-packages"""
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
six = "*"

[dev-packages]
pytest = "*"
            """.strip()
            f.write(contents)

        # Install both packages
        c = p.pipenv("install --dev")
        assert c.returncode == 0
        assert "six" in p.pipfile["packages"]
        assert "pytest" in p.pipfile["dev-packages"]
        assert "six" in p.lockfile["default"]
        assert "pytest" in p.lockfile["develop"]

        # Verify both packages are installed
        c = p.pipenv('run python -c "import six, pytest"')
        assert c.returncode == 0

        # Uninstall pytest with --dev flag
        c = p.pipenv("uninstall pytest --dev")
        assert c.returncode == 0

        # Verify pytest was removed from dev-packages
        assert "six" in p.pipfile["packages"]
        assert "pytest" not in p.pipfile["dev-packages"]
        assert "six" in p.lockfile["default"]
        assert "pytest" not in p.lockfile["develop"]

        # Verify pytest is no longer importable
        c = p.pipenv('run python -c "import pytest"')
        assert c.returncode != 0

        # Verify six is still importable
        c = p.pipenv('run python -c "import six"')
        assert c.returncode == 0


@pytest.mark.install
@pytest.mark.uninstall
def test_uninstall_dev_flag_with_categories(pipenv_instance_private_pypi):
    """Ensure that running `pipenv uninstall --dev` works the same as `--categories dev-packages`"""
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
six = "*"

[dev-packages]
pytest = "*"
            """.strip()
            f.write(contents)

        # Install both packages
        c = p.pipenv("install --dev")
        assert c.returncode == 0

        # Create a second project to test with categories
        with pipenv_instance_private_pypi() as p2:
            with open(p2.pipfile_path, "w") as f:
                contents = """
[packages]
six = "*"

[dev-packages]
pytest = "*"
                """.strip()
                f.write(contents)

            # Install both packages
            c = p2.pipenv("install --dev")
            assert c.returncode == 0

            # Uninstall pytest with --categories
            c = p2.pipenv("uninstall pytest --categories dev-packages")
            assert c.returncode == 0

            # Verify pytest was removed from dev-packages
            assert "six" in p2.pipfile["packages"]
            assert "pytest" not in p2.pipfile["dev-packages"]
            assert "six" in p2.lockfile["default"]
            assert "pytest" not in p2.lockfile["develop"]

            # Compare with first project
            c = p.pipenv("uninstall pytest --dev")
            assert c.returncode == 0

            # Verify both approaches have the same result
            assert p.pipfile["packages"] == p2.pipfile["packages"]
            assert p.pipfile["dev-packages"] == p2.pipfile["dev-packages"]
            assert p.lockfile["default"] == p2.lockfile["default"]
            assert p.lockfile["develop"] == p2.lockfile["develop"]



import pytest

from pipenv.utils.virtualenv import ensure_python


def test_ensure_python_non_interactive_no_yes(monkeypatch, project):
    """Test ensure_python when SESSION_IS_INTERACTIVE=False and PIPENV_YES=False."""
    # Mock the environments.SESSION_IS_INTERACTIVE to be False
    monkeypatch.setattr("pipenv.environments.SESSION_IS_INTERACTIVE", False)

    # Mock project.s.PIPENV_YES to be False
    monkeypatch.setattr(project.s, "PIPENV_YES", False)

    # Mock find_version_to_install to return a version
    class MockInstaller:
        def __init__(self, *args, **kwargs):
            self.cmd = "mock_installer"

        def find_version_to_install(self, *args, **kwargs):
            return "3.11.0"

    monkeypatch.setattr("pipenv.installers.Pyenv", MockInstaller)

    # Mock find_a_system_python to return None (Python not found)
    monkeypatch.setattr("pipenv.utils.virtualenv.find_a_system_python", lambda x: None)

    # Mock os.name to not be 'nt' to skip Windows-specific code
    monkeypatch.setattr("os.name", "posix")

    # Mock project.s.PIPENV_DONT_USE_PYENV to be False
    monkeypatch.setattr(project.s, "PIPENV_DONT_USE_PYENV", False)

    # The function should call sys.exit(1) when SESSION_IS_INTERACTIVE=False and PIPENV_YES=False
    # We'll catch this with pytest.raises
    with pytest.raises(SystemExit) as excinfo:
        ensure_python(project, python="3.11.0")

    # Verify that sys.exit was called with code 1
    assert excinfo.value.code == 1


def test_ensure_python_non_interactive_with_yes(monkeypatch, project):
    """Test ensure_python when SESSION_IS_INTERACTIVE=False but PIPENV_YES=True."""
    # Mock the environments.SESSION_IS_INTERACTIVE to be False
    monkeypatch.setattr("pipenv.environments.SESSION_IS_INTERACTIVE", False)

    # Mock project.s.PIPENV_YES to be True
    monkeypatch.setattr(project.s, "PIPENV_YES", True)

    # Mock find_version_to_install to return a version
    class MockInstaller:
        def __init__(self, *args, **kwargs):
            self.cmd = "mock_installer"

        def find_version_to_install(self, *args, **kwargs):
            return "3.11.0"

        def install(self, *args, **kwargs):
            class Result:
                stdout = "Installed successfully"
            return Result()

    monkeypatch.setattr("pipenv.installers.Pyenv", MockInstaller)

    # Mock find_a_system_python to return None initially (Python not found)
    # and then return a path after "installation"
    find_python_calls = [None]

    def mock_find_python(version):
        if len(find_python_calls) == 1:
            find_python_calls.append("/mock/path/to/python")
            return find_python_calls[-1]
        return find_python_calls[-1]

    monkeypatch.setattr("pipenv.utils.virtualenv.find_a_system_python", mock_find_python)

    # Mock python_version to return the expected version
    monkeypatch.setattr("pipenv.utils.dependencies.python_version", lambda x: "3.11.0")

    # Mock os.name to not be 'nt' to skip Windows-specific code
    monkeypatch.setattr("os.name", "posix")

    # Mock project.s.PIPENV_DONT_USE_PYENV to be False
    monkeypatch.setattr(project.s, "PIPENV_DONT_USE_PYENV", False)

    # Mock console.status to do nothing
    def mock_status(*args, **kwargs):
        class MockContextManager:
            def __enter__(self):
                return None

            def __exit__(self, *args):
                pass

        return MockContextManager()

    monkeypatch.setattr("pipenv.utils.console.status", mock_status)

    # The function should proceed with installation when PIPENV_YES=True
    result = ensure_python(project, python="3.11.0")

    # Verify that the function returned the path to Python
    assert result == "/mock/path/to/python"



import os
import sys

import pytest


@pytest.mark.categories
@pytest.mark.install
def test_basic_category_install(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        c = p.pipenv("install six --categories prereq")
        assert c.returncode == 0
        assert "six" not in p.pipfile["packages"]
        assert "six" not in p.lockfile["default"]
        assert "six" in p.pipfile["prereq"]
        assert "six" in p.lockfile["prereq"]


@pytest.mark.categories
@pytest.mark.install
@pytest.mark.requirements
def test_basic_category_install_from_requirements(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi(pipfile=False) as p:
        # Write a requirements file
        with open("requirements.txt", "w") as f:
            f.write("six==1.16.0")

        c = p.pipenv("install --categories prereq")
        assert c.returncode == 0
        os.unlink("requirements.txt")
        print(c.stdout)
        print(c.stderr)
        # assert stuff in pipfile
        assert c.returncode == 0
        assert "six" not in p.pipfile["packages"]
        assert "six" not in p.lockfile["default"]
        assert "six" in p.pipfile["prereq"]
        assert "six" in p.lockfile["prereq"]


@pytest.mark.categories
@pytest.mark.install
@pytest.mark.parametrize("categories", ["prereq other", "prereq, other"])
def test_multiple_category_install(pipenv_instance_private_pypi, categories):
    with pipenv_instance_private_pypi() as p:
        c = p.pipenv('install six --categories="prereq other"')
        assert c.returncode == 0
        assert "six" not in p.pipfile["packages"]
        assert "six" not in p.lockfile["default"]
        assert "six" in p.pipfile["prereq"]
        assert "six" in p.lockfile["prereq"]
        assert "six" in p.lockfile["other"]
        assert "six" in p.lockfile["other"]


@pytest.mark.categories
@pytest.mark.install
@pytest.mark.requirements
def test_multiple_category_install_from_requirements(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi(pipfile=False) as p:
        # Write a requirements file
        with open("requirements.txt", "w") as f:
            f.write("six==1.16.0")

        c = p.pipenv('install --categories="prereq other"')
        assert c.returncode == 0
        os.unlink("requirements.txt")
        print(c.stdout)
        print(c.stderr)
        # assert stuff in pipfile
        assert c.returncode == 0
        assert "six" not in p.pipfile["packages"]
        assert "six" not in p.lockfile["default"]
        assert "six" in p.pipfile["prereq"]
        assert "six" in p.lockfile["prereq"]
        assert "six" in p.pipfile["other"]
        assert "six" in p.lockfile["other"]


@pytest.mark.extras
@pytest.mark.install
@pytest.mark.local
@pytest.mark.skipif(sys.version_info >= (3, 12), reason="test is not 3.12 compatible")
def test_multiple_category_install_proceeds_in_order_specified(
    pipenv_instance_private_pypi,
):
    """Ensure -e .[extras] installs."""
    with pipenv_instance_private_pypi() as p:
        setup_py = os.path.join(p.path, "setup.py")
        with open(setup_py, "w") as fh:
            contents = """
import six
from setuptools import setup
setup(
    name='testpipenv',
    version='0.1',
    description='Pipenv Test Package',
    author='Pipenv Test',
    author_email='test@pipenv.package',
    license='MIT',
    packages=[],
    install_requires=['six', 'setuptools'],
    zip_safe=False
)
            """.strip()
            fh.write(contents)
        with open(os.path.join(p.path, "Pipfile"), "w") as fh:
            fh.write(
                """
[packages]
testpipenv = {path = ".", editable = true, skip_resolver = true}

[prereq]
six = "*"
            """.strip()
            )
        c = p.pipenv("lock -v")
        assert c.returncode == 0
        assert "testpipenv" in p.lockfile["default"]
        assert "testpipenv" not in p.lockfile["prereq"]
        assert "six" in p.lockfile["prereq"]
        c = p.pipenv(
            'sync --categories="prereq packages" --extra-pip-args="--no-build-isolation" -v'
        )
        assert c.returncode == 0



import pytest

from pipenv.project import Project
from pipenv.routines.update import get_modified_pipfile_entries


@pytest.mark.parametrize("cmd_option", ["", "--dev"])
@pytest.mark.basic
@pytest.mark.update
@pytest.mark.skipif(
    "os.name == 'nt' and sys.version_info[:2] == (3, 8)",
    reason="Seems to work on 3.8 but not via the CI",
)
def test_update_outdated_with_outdated_package(pipenv_instance_private_pypi, cmd_option):
    with pipenv_instance_private_pypi() as p:
        package_name = "six"
        p.pipenv(f"install {cmd_option} {package_name}==1.11")
        c = p.pipenv(f"update {package_name} {cmd_option} --outdated")
        assert f"Package '{package_name}' out-of-date:" in c.stdout


def test_get_modified_pipfile_entries_new_package(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        # Add package to Pipfile
        p.pipenv("install requests==2.31.0")

        # Add new package to Pipfile
        pipfile = p.pipfile_path
        content = pipfile.read_text()
        content = content.replace(
            '[packages]',
            '[packages]\nurllib3 = "==2.0.0"'
        )
        pipfile.write_text(content)
        project = Project()
        modified = get_modified_pipfile_entries(project, ["packages"])
        assert "urllib3" in modified["default"]


def test_get_modified_pipfile_entries_changed_version(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        # Install and lock initial version
        p.pipenv("install requests==2.31.0")

        # Update version in Pipfile
        pipfile = p.pipfile_path
        content = pipfile.read_text()
        content = content.replace(
            'requests = "==2.31.0"',
            'requests = "==2.32.0"'
        )
        pipfile.write_text(content)
        project = Project()
        modified = get_modified_pipfile_entries(project, ["packages"])
        assert "requests" in modified["default"]


def test_get_modified_pipfile_entries_vcs_changes(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        # Install VCS package
        p.pipenv("install git+https://github.com/requests/requests.git@main#egg=requests")

        # Change ref
        pipfile = p.pipfile_path
        content = pipfile.read_text()
        content = content.replace(
            'ref = "main"',
            'ref = "master"'
        )
        pipfile.write_text(content)
        project = Project()
        modified = get_modified_pipfile_entries(project, ["packages"])
        assert "requests" in modified["default"]


def test_update_without_lockfile(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        # Install without creating lockfile
        p.pipenv("install --skip-lock requests==2.31.0")

        # Update should work without existing lockfile
        c = p.pipenv("update requests")
        assert c.returncode == 0


@pytest.mark.parametrize(
    "initial_content,modified_content,expected_updates",
    [
        # Test no changes case
        (
            '[packages]\nrequests = "==2.31.0"',
            '[packages]\nrequests = "==2.31.0"',
            set()
        ),
        # Test version change
        (
            '[packages]\nrequests = "==2.31.0"',
            '[packages]\nrequests = "==2.32.0"',
            {"requests"}
        ),
        # Test multiple package changes
        (
            '[packages]\nrequests = "==2.31.0"',
            '[packages]\nrequests = "==2.32.0"\nurllib3 = "==2.0.0"',
            {"requests", "urllib3"}
        ),
        # Test dev packages
        (
            '[packages]\nrequests = "==2.31.0"\n[dev-packages]',
            '[packages]\nrequests = "==2.31.0"\n[dev-packages]\npytest = "==7.4.0"',
            {"pytest"}
        ),
        # Test VCS package
        (
            '[packages]\nrequests = {git = "https://github.com/requests/requests.git", ref = "main"}',
            '[packages]\nrequests = {git = "https://github.com/requests/requests.git", ref = "v2.31.0"}',
            {"requests"}
        ),
        # Test extras change
        (
            '[packages]\nrequests = {version = "==2.31.0", extras = ["security"]}',
            '[packages]\nrequests = {version = "==2.31.0", extras = ["security", "socks"]}',
            {"requests"}
        )
    ]
)
def test_update_modified_packages(pipenv_instance_pypi, initial_content, modified_content, expected_updates):
    with pipenv_instance_pypi() as p:
        # Write initial Pipfile
        p.pipfile_path.write_text(initial_content)
        p.pipenv("lock")  # Generate initial lockfile

        # Modify Pipfile
        p.pipfile_path.write_text(modified_content)

        project = Project()
        # Verify correct packages identified for update
        modified = get_modified_pipfile_entries(project, ["packages", "dev-packages"])
        all_modified = set()
        for category in modified.values():
            all_modified.update(entry if isinstance(entry, str) else entry.get("name", "") for entry in category)

        assert all_modified == expected_updates



import json
import os

import pytest


@pytest.mark.upgrade
def test_category_sorted_alphabetically_with_directive(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[pipenv]
sort_pipfile = true

[packages]
zipp = "*"
six = 1.11
colorama = "*"
atomicwrites = "*"
            """.strip()
            f.write(contents)

        package_name = "six"
        c = p.pipenv(f"upgrade {package_name}")
        assert c.returncode == 0
        assert list(p.pipfile["packages"].keys()) == [
            "atomicwrites",
            "colorama",
            "six",
            "zipp",
        ]


@pytest.mark.upgrade
def test_category_not_sorted_without_directive(pipenv_instance_private_pypi):
    with pipenv_instance_private_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
zipp = "*"
six = 1.11
colorama = "*"
atomicwrites = "*"
            """.strip()
            f.write(contents)

        package_name = "six"
        c = p.pipenv(f"upgrade {package_name}")
        assert c.returncode == 0
        assert list(p.pipfile["packages"].keys()) == [
            "zipp",
            "colorama",
            "atomicwrites",
            "six",
        ]


@pytest.mark.cli
def test_pipenv_dependency_incompatibility_resolution(pipenv_instance_pypi):
    from pipenv.cli import cli
    from pipenv.vendor.click.testing import CliRunner

    with pipenv_instance_pypi() as p:
        # Step 1: Install initial dependency version
        c = p.pipenv("install google-api-core==2.18.0")
        assert c.returncode == 0, f"Failed to install google-api-core: {c.stderr}"

        # Ensure initial state
        lockfile_path = os.path.join(p.path, "Pipfile.lock")
        with open(lockfile_path) as lockfile:
            lock_data = json.load(lockfile)
        assert "google-api-core" in lock_data["default"]
        assert lock_data["default"]["google-api-core"]["version"] == "==2.18.0"

        # Step 2: Update Pipfile to allow any version of google-api-core
        pipfile_path = os.path.join(p.path, "Pipfile")
        with open(pipfile_path) as pipfile:
            pipfile_content = pipfile.read()

        updated_pipfile_content = pipfile_content.replace("google-api-core = \"==2.18.0\"", "google-api-core = \"*\"")
        with open(pipfile_path, "w") as pipfile:
            pipfile.write(updated_pipfile_content)

        # Step 3: Update protobuf to an incompatible version
        cli_runner = CliRunner()
        c = cli_runner.invoke(cli, "update protobuf==5.27.5")
        assert c.exit_code == 0, f"Failed to update protobuf: {c.stderr}"

        # Step 4: Check the lockfile for incompatible dependencies
        with open(lockfile_path) as lockfile:
            lock_data = json.load(lockfile)

        # Check if google-api-core is still at the old version
        google_api_core_version = lock_data["default"].get("google-api-core", {}).get("version", "")
        protobuf_version = lock_data["default"].get("protobuf", {}).get("version", "")

        assert google_api_core_version != "==2.18.0", (
            "google-api-core was not updated to a compatible version despite the protobuf update"
        )
        assert protobuf_version == "==5.27.5", "protobuf was not updated correctly"

        # Step 5: Run pipenv lock to check for dependency resolution errors
        c = cli_runner.invoke(cli, "lock")
        assert c.exit_code == 0, f"Failed to run pipenv lock: {c.stderr}"


@pytest.mark.upgrade
def test_upgrade_updates_lockfile_in_all_categories(pipenv_instance_private_pypi):
    """Test that upgrading a package updates it in all categories of the lockfile."""
    with pipenv_instance_private_pypi() as p:
        # Create a Pipfile with a package in default and a dev package that depends on it
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
requests = "==2.25.0"

[dev-packages]
pytest = "*"
            """.strip()
            f.write(contents)

        # Lock the dependencies
        c = p.pipenv("lock")
        assert c.returncode == 0, f"Failed to lock dependencies: {c.stderr}"

        # Verify initial state
        lockfile_path = os.path.join(p.path, "Pipfile.lock")
        with open(lockfile_path) as lockfile:
            lock_data = json.load(lockfile)

        # Check initial version in default section
        assert lock_data["default"]["requests"]["version"] == "==2.25.0"

        # Check if requests is in develop section (as a dependency of pytest)
        develop_has_requests = "requests" in lock_data["develop"]

        # Upgrade the package
        target_version = "2.28.0"
        c = p.pipenv(f"upgrade requests=={target_version}")
        assert c.returncode == 0, f"Failed to upgrade requests: {c.stderr}"

        # Verify the package was updated in the lockfile
        with open(lockfile_path) as lockfile:
            updated_lock_data = json.load(lockfile)

        # Check updated version in default section
        assert updated_lock_data["default"]["requests"]["version"] == f"=={target_version}"

        # If requests was in develop section, check it was updated there too
        if develop_has_requests:
            assert updated_lock_data["develop"]["requests"]["version"] == f"=={target_version}"


@pytest.mark.upgrade
def test_upgrade_only_adds_to_explicit_categories(pipenv_instance_private_pypi):
    """Test that upgrading a package only adds it to the Pipfile for explicitly requested categories."""
    with pipenv_instance_private_pypi() as p:
        # Create a Pipfile with a package in default but not in dev-packages
        with open(p.pipfile_path, "w") as f:
            contents = """
[packages]
requests = "==2.25.0"

[dev-packages]
pytest = "*"
            """.strip()
            f.write(contents)

        # Lock the dependencies
        c = p.pipenv("lock")
        assert c.returncode == 0, f"Failed to lock dependencies: {c.stderr}"

        # Verify initial state
        with open(p.pipfile_path) as pipfile:
            initial_pipfile_content = pipfile.read()

        # Make sure requests is not in dev-packages initially
        assert "requests" not in initial_pipfile_content.split("[dev-packages]")[1].split("[")[0]

        # Upgrade the package
        target_version = "2.28.0"
        c = p.pipenv(f"upgrade requests=={target_version}")
        assert c.returncode == 0, f"Failed to upgrade requests: {c.stderr}"

        # Verify the Pipfile was updated correctly
        with open(p.pipfile_path) as pipfile:
            updated_pipfile_content = pipfile.read()

        # Check that requests was updated in packages section
        packages_section = updated_pipfile_content.split("[packages]")[1].split("[")[0]
        assert f'requests = "=={target_version}"' in packages_section

        # Check that requests was NOT added to dev-packages section
        if "[dev-packages]" in updated_pipfile_content:
            dev_packages_section = updated_pipfile_content.split("[dev-packages]")[1].split("[")[0]
        else:
            dev_packages_section = ""
        assert "requests" not in dev_packages_section

        # Verify the lockfile was updated in both sections
        lockfile_path = os.path.join(p.path, "Pipfile.lock")
        with open(lockfile_path) as lockfile:
            updated_lock_data = json.load(lockfile)

        # Check that requests was updated in default section
        assert updated_lock_data["default"]["requests"]["version"] == f"=={target_version}"

        # Check if requests is in develop section (it might be there as a dependency of pytest)
        if "requests" in updated_lock_data["develop"]:
            # If it's there, it should be updated
            assert updated_lock_data["develop"]["requests"]["version"] == f"=={target_version}"



import pytest

from pipenv.utils.shell import temp_environ


@pytest.mark.lock
@pytest.mark.sync
def test_sync_error_without_lockfile(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[packages]
            """.strip()
            )

        c = p.pipenv("sync")
        assert c.returncode != 0
        assert "Pipfile.lock not found!" in c.stderr


@pytest.mark.sync
@pytest.mark.lock
def test_mirror_lock_sync(pipenv_instance_private_pypi):
    with temp_environ(), pipenv_instance_private_pypi() as p:
        mirror_url = p.index_url
        assert "pypi.org" not in mirror_url
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[[source]]
name = "pypi"
url = "https://pypi.org/simple"
verify_ssl = true

[packages]
six = "==1.12.0"
            """.strip()
            )
        c = p.pipenv(f"lock --pypi-mirror {mirror_url}")
        assert c.returncode == 0
        c = p.pipenv(f"sync --pypi-mirror {mirror_url}")
        assert c.returncode == 0


@pytest.mark.sync
@pytest.mark.lock
def test_sync_should_not_lock(pipenv_instance_pypi):
    """Sync should not touch the lock file, even if Pipfile is changed."""
    with pipenv_instance_pypi() as p:
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[packages]
            """.strip()
            )

        # Perform initial lock.
        c = p.pipenv("lock")
        assert c.returncode == 0
        lockfile_content = p.lockfile
        assert lockfile_content

        # Make sure sync does not trigger lockfile update.
        with open(p.pipfile_path, "w") as f:
            f.write(
                """
[packages]
six = "*"
            """.strip()
            )
        c = p.pipenv("sync")
        assert c.returncode == 0
        assert lockfile_content == p.lockfile



import os
from pathlib import Path

import pytest


@pytest.mark.install
@pytest.mark.needs_internet
@pytest.mark.skipif(
    os.name == "nt",
    reason="This test is not for Windows",
)
def test_install_path_with_nested_setup_module(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        # Install setuptools first for Python 3.12+ compatibility
        c = p.pipenv("install setuptools")
        assert c.returncode == 0

        # Create a simple package.
        package_dir = Path(p.path) / "simple_package"
        package_dir.mkdir()

        # Create a simple setup.py file
        with open(package_dir / "setup.py", "w") as f:
            f.write("""
from setuptools import setup, find_packages

setup(
    name="simple-package",
    version="0.1.0",
    packages=find_packages(),
)
""")
        module_dir = package_dir / "simple_package"
        module_dir.mkdir()

        (module_dir / "__init__.py").touch()

        # Create a test module that contains setup.py-like content but isn't setup.py
        with open(module_dir / "test_setup.py", "w") as f:
            f.write("""
def setup(name = ''):
    return f'Setup package {name}'

def configure():
    return setup(name='my_simple_package')
""")

        # Install the package using a path with spaces
        # Use both escaped spaces and quoted path to test both scenarios
        relative_package_dir = f'"{package_dir.relative_to(p.path)}"'
        c = p.pipenv(f'install -e {relative_package_dir}')
        assert c.returncode == 0

        # Verify the package was installed correctly
        c = p.pipenv('run python -c "'
                     'import simple_package.test_setup; '
                     'print(simple_package.test_setup.configure())"')
        assert c.returncode == 0
        assert "Setup package my_simple_package" in c.stdout

        # Ensure my_simple_package was not parsed from the test_setup.py module
        # inside the package - this is the key test for our fix.
        with open(Path(p.path) / "Pipfile") as f:
            pipfile_content = f.read()
            assert "my_simple_package" not in pipfile_content
            assert "simple-package" in pipfile_content  # Should find the correct name



import os
from pathlib import Path

import pytest


@pytest.mark.install
@pytest.mark.needs_internet
@pytest.mark.skipif(
    os.name == "nt",
    reason="This test is not for Windows",
)
def test_install_path_with_spaces(pipenv_instance_pypi):
    with pipenv_instance_pypi() as p:
        # Create a test directory with spaces in the name
        test_dir = Path(p.path) / "test dir with spaces"
        test_dir.mkdir()

        # Create a simple package in the directory with spaces
        package_dir = test_dir / "simple_package"
        package_dir.mkdir()

        # Create a simple setup.py file
        with open(package_dir / "setup.py", "w") as f:
            f.write("""
from setuptools import setup

setup(
    name="simple-package",
    version="0.1.0",
)
""")

        # Create the package module
        with open(package_dir / "simple_package.py", "w") as f:
            f.write("""
def hello():
    return "Hello from simple package!"
""")

        # Install the package using a path with spaces
        # Use both escaped spaces and quoted path to test both scenarios
        c = p.pipenv(f'install "{package_dir}"')
        assert c.returncode == 0

        # Verify the package was installed correctly
        c = p.pipenv('run python -c "import simple_package; print(simple_package.hello())"')
        assert c.returncode == 0
        assert "Hello from simple package!" in c.stdout

        # Test with escaped spaces
        p.pipenv("uninstall simple-package")
        escaped_path = package_dir.as_posix().replace(" ", "\\ ")
        c = p.pipenv(f'install {escaped_path}')
        assert c.returncode == 0

        # Verify the package was installed correctly
        c = p.pipenv('run python -c "import simple_package; print(simple_package.hello())"')
        assert c.returncode == 0
        assert "Hello from simple package!" in c.stdout
